/*
 * mkstorage.cpp --
 *
 *	This file contains the implementation of the 
 *	e4_MetakitStorageImpl class.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Include the e4Graph types used directly here.
 */

#include "e4graphimpl.h"

/*
 * Include the type information specific to this storage driver.
 */

#include "mkstorage.h"

/*
 * These constants define bits for recording state information about
 * an entity (vertex or node) in a storage.
 *
 * E4_INUSE			Persistent, notes that the entity is
 *				in use.
 * E4_REACHABLE			Transient, used during GC to note
 *				that this entity is reachable.
 * E4_DETACHED			Persistent, notes that this entity
 *				is detached. It will become unreachable
 *				if the user program drops the last
 *				reference to it.
 * E4_DETACHNOTIFY		Transient, notes that this detached
 *				entity has had a detach notification
 *				delivered to it.
 */

#define E4_INUSE			(1<<0)
#define E4_REACHABLE			(1<<1)
#define E4_DETACHED			(1<<2)
#define E4_DETACHNOTIFY			(1<<3)

/*
 * Properties for indexing entities in a storage (common across 1.0, 1.1,
 * 1.2 and 1.3):
 */

static c4_IntProp pUserData("userData");
static c4_IntProp pNameID("nameID");
static c4_IntProp pNext("next");
static c4_IntProp pPrev("prev");

static c4_ViewProp pMarkers("e4GraphMarkers");
  static c4_IntProp pRoot("root");

static c4_ViewProp pNodes("e4GraphNodes");
  static c4_IntProp pFirstVertex("firstVertex");
  static c4_IntProp pLastVertex("lastVertex");
  static c4_IntProp pParentID("parentID");
  static c4_IntProp pVertexCount("vertexCount");
  static c4_IntProp pRefCount("refCount");
  static c4_IntProp pNodeMarkers("nodeMarkers");

static c4_ViewProp pVertices("e4GraphVertices");
  static c4_IntProp pNodeID("nodeID");
  static c4_IntProp pVertexType("vertexType");
  static c4_IntProp pRowID("row");

static c4_ViewProp pInts("e4GraphInts");
  static c4_IntProp pIntVal("i");

static c4_ViewProp pDoubles("e4GraphDoubles");
  static c4_DoubleProp pDoubleVal("d");

static c4_ViewProp pStrings("e4GraphStrings");
  static c4_StringProp pStringVal("s");

static c4_ViewProp pNames("e4GraphNames");
  static c4_StringProp pNameVal("n");

static c4_ViewProp pBinary("e4Graphbinary");
  static c4_BytesProp pBinaryVal("b");

static c4_ViewProp pUnused("e4GraphUnused");
  static c4_IntProp pFirst("first");

static c4_ViewProp pParents("e4GraphParents");
  static c4_IntProp pCount("count");

/*
 * In 1.4 the marker properties are defined but unused. We need them
 * to be defined to be able to convert storages written in an older
 * version to the newer format.
 */

/*
 * Properties present only since 1.3:
 */

/*
 * The "detachedvertices" property (occurs in nodes) points at the head
 * of a list of detached vertices that have this node as their value. These
 * detached vertices are chained through their "nextinparent" property.
 */

static c4_IntProp pDetachedVertices("detachedvertices");

/*
 * Properties present only since 1.2:
 */

/*
 * The "vertexchain" property (occurs in parent records) contains the start
 * of a chain of vertices in a parent node that have a specific child node
 * as their value.
 */

static c4_IntProp pVertexChain("vertexchain");

/*
 * The "nextinparent" property (occurs in vertices) contains the next vertex
 * in a chain of vertices.
 */

static c4_IntProp pNextInParent("nextinparent");

/*
 * Properties present only since 1.1:
 */

/*
 * The "flags" property (occurs on all items in a storage) replaces and
 * generalizes the "used" property that was removed since 1.1.
 */

static c4_IntProp pFlags("flags");

/*
 * Properties present only in 1.0, removed later:
 */

static c4_IntProp pUsed("used");

/*
 * This class static method is used to create a new
 * instance of e4_MetakitStorageImpl or return an existing one.
 */

e4_StorageImpl *
e4_MetakitStorageImpl::GetStorage(const char *fname)
{
    return new e4_MetakitStorageImpl(fname);
}

/*
 * This constructor opens a storage in write mode.
 */

e4_MetakitStorageImpl::e4_MetakitStorageImpl(const char *fname) 
    : e4_StorageImpl(fname, E4_METAKIT)
{
    /*
     * Open the storage in exclusive read-write mode.
     */

    storage = new c4_Storage(fname, 1);

    (void) Initialize();
}

/*
 * Helper function for initializing a storage.
 */

bool
e4_MetakitStorageImpl::Initialize()
{
    int i, j, cnt, flags;

    unused = storage->GetAs(E4_GRAPHUNUSED1_4);
    if (unused.GetSize() == 0) {
        unused.SetSize(E4_GRAPHUNUSEDSIZE);
	pFirst(unused[E4_GRAPHSTORAGEMAJORVER]) = (int) E4_MKSTORAGE_MAJORVER;
	pFirst(unused[E4_GRAPHSTORAGEMINORVER]) = (int) E4_MKSTORAGE_MINORVER;
        for (i = E4_GRAPHFIRSTUNUSEDMARKER; i < E4_GRAPHUNUSEDSIZE; i++) {
            pFirst(unused[i]) = (int) E4_NEXTNONE;
        }

	/*
	 * Need better source of randomness than the address of the storage!
	 */

	pFirst(unused[E4_GRAPHHASHCODE]) = (int) this;

	/*
	 * The storage is initially clean -- no leaked data.
	 */

	pFirst(unused[E4_GRAPHUNREACHABLENODES]) = 0;

	/*
	 * Initially we do not defer GC.
	 */

	pFirst(unused[E4_GRAPHDEFERGC]) = 0;

	/*
	 * Initialize the root node.
	 */

	pFirst(unused[E4_GRAPHROOTNODE]) = E4_NEXTNONE;
    }

    /*
     * Ensure the data stored in the storage just opened is in a
     * format compatible with the current version of the driver.
     */

    if ((int) pFirst(unused[E4_GRAPHSTORAGEMAJORVER]) != 
	E4_MKSTORAGE_MAJORVER) {
	fprintf(stderr, 
		"**Warning**: version mismatch: storage version %d.%d, ",
		(int) pFirst(unused[E4_GRAPHSTORAGEMAJORVER]),
		(int) pFirst(unused[E4_GRAPHSTORAGEMINORVER]));
	fprintf(stderr, 
		"expected version: %d.X. ABORTING\n", 
		E4_MKSTORAGE_MAJORVER);
	return false;
    }

    /*
     * Check if the minor version is older than what we expect. If it is
     * then fix the storage transparently to be of the format expected by
     * the newer minor version, if possible.
     */

    switch ((int) pFirst(unused[E4_GRAPHSTORAGEMINORVER])) {
    case 0:
	/*
	 * Attempt to convert format 1.0 to format 1.1:
	 */

	UpdateFormat1_0to1_1();

	/*
	 * Fall through:
	 */

    case 1:

	/*
	 * Attempt to convert format 1.1 to format 1.2:
	 */

	UpdateFormat1_1to1_2();

	/*
	 * Fall through:
	 */

    case 2:

	UpdateFormat1_2to1_3();

	/*
	 * Fall through:
	 */

    case 3:

	/*
	 * Attempt to convert format 1.3 to format 1.4:
	 */

	UpdateFormat1_3to1_4();

	/*
	 * Fall through:
	 */

    case 4:

	/*
	 * Get the storage structure as expected by the current minor
	 * version.
	 */

	nodes = storage->GetAs(E4_GRAPHNODES1_4);
	vertices = storage->GetAs(E4_GRAPHVERTICES1_4);
	doubles = storage->GetAs(E4_GRAPHDOUBLES1_4);
	strings = storage->GetAs(E4_GRAPHSTRINGS1_4);
	binary = storage->GetAs(E4_GRAPHBINARY1_4);
	names = storage->GetAs(E4_GRAPHNAMES1_4);
	parents = storage->GetAs(E4_GRAPHPARENTS1_4);
	break;

    default:
	fprintf(stderr,
		"Expected version %d.%d or earlier, got %d.%d. ABORTING\n",
		E4_MKSTORAGE_MAJORVER, E4_MKSTORAGE_MINORVER,
		E4_MKSTORAGE_MAJORVER, 
		(int) pFirst(unused[E4_GRAPHSTORAGEMINORVER]));
	return false;
    }

    /*
     * Hash all the names.
     */

    PopulateNameHash();

    for (cnt = 0, i = 0; i < nodes.GetSize(); i++) {
      if (((int) pFlags(nodes[i]) & E4_INUSE) == E4_INUSE) {
	cnt++;
      }
    }

    /*
     * Initialize the array of statistics.
     */

    for (i = 0; i < (int) E4_SPLAST; i++) {
	for (j = 0; j < (int) E4_SSLAST; j++) {
	    statistics[i][j] = 0;
	}
    }
    statistics[(int) E4_SPNODE][(int) E4_SSAVAIL] = nodes.GetSize();
    statistics[(int) E4_SPVERTEX][(int) E4_SSAVAIL] = vertices.GetSize();
    statistics[(int) E4_SPNAME][(int) E4_SSAVAIL] = names.GetSize();
    statistics[(int) E4_SPSTRING][(int) E4_SSAVAIL] = strings.GetSize();
    statistics[(int) E4_SPDOUBLE][(int) E4_SSAVAIL] = doubles.GetSize();
    statistics[(int) E4_SPBINARY][(int) E4_SSAVAIL] = binary.GetSize();

    for (i = 0, cnt = 0; i < nodes.GetSize(); i++) {
      if (((int) pFlags(nodes[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPNODE][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < vertices.GetSize(); i++) {
	if (((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPVERTEX][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < names.GetSize(); i++) {
	if (((int) pFlags(names[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPNAME][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < strings.GetSize(); i++) {
	if (((int) pFlags(strings[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPSTRING][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < doubles.GetSize(); i++) {
	if (((int) pFlags(doubles[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPDOUBLE][(int) E4_SSUSED] = cnt;
    for (i = 0, cnt = 0; i < binary.GetSize(); i++) {
	if (((int) pFlags(binary[i]) & E4_INUSE) == E4_INUSE) {
	    cnt++;
	}
    }
    statistics[(int) E4_SPBINARY][(int) E4_SSUSED] = cnt;

    /*
     * Initialize the deletion stacks.
     */

    idStack1 = new e4_IntStack();
    idStack2 = new e4_IntStack();

    /*
     * Initialize the GC counters and boolean needsGC indicator.
     */

    needsGC = false;

    /*
     * Clean up any remaining detached entities after the last commit.
     */

    CleanupDetached();

    /*
     * If there's no root node, reserve it now.
     */

    if ((int) pFirst(unused[E4_GRAPHROOTNODE]) == E4_NEXTNONE) {
	pFirst(unused[E4_GRAPHROOTNODE]) = i = ReserveNodeID();

	/*
	 * Special handling -- we do not want to fire a detach
	 * event on the root node we just created, so mark it
	 * as already having had an event fired on it.
	 */

	flags = (int) pFlags(nodes[i]);
	flags |= E4_DETACHNOTIFY;
	pFlags(nodes[i]) = flags;
    }

    return true;
}

/*
 * Clean up any remaining detached entities after the last commit. This
 * is done when a storage is re-opened.
 *
 * NOTE: We just find the first detached entity and delete it. The GC
 * mechanism will find all other detached unreachable entities.
 */

void
e4_MetakitStorageImpl::CleanupDetached()
{
    int i, l, r, flags;

    /*
     * Search for any detached vertices.
     */

    for (i = 0, l = vertices.GetSize(); i < l; i++) {
	flags = (int) pFlags(vertices[i]);
	if ((flags & (E4_INUSE | E4_DETACHED)) == (E4_INUSE | E4_DETACHED)) {
	    flags &= ~(E4_DETACHED | E4_DETACHNOTIFY);
	    pFlags(vertices[i]) = flags;
	    if (!IsGCDeferred()) {
		DoGC();
	    } else {
		needsGC = true;
	    }
	    return;
	}
    }

    /*
     * Search for any detached nodes. Skip the root node because it
     * might be detached but is still reachable.
     */

    for (i = 0, r = (int) pFirst(unused[E4_GRAPHROOTNODE]),
	   l = nodes.GetSize(); 
	 i < l;
	 i++) {

	/*
	 * Skip the root node.
	 */

	if (i == r) {
	    continue;
	}

	/*
	 * Check if this node is detached.
	 */

	flags = (int) pFlags(nodes[i]);
	if ((flags & (E4_INUSE | E4_DETACHED)) == (E4_INUSE | E4_DETACHED)) {
	    flags &= ~(E4_DETACHED | E4_DETACHNOTIFY);
	    pFlags(nodes[i]) = flags;
	    if (!IsGCDeferred()) {
		DoGC();
	    } else {
		needsGC = true;
	    }
	    return;
	}
    }
}

/*
 * Helper method to update the format from 1.3 to 1.4:
 */

void
e4_MetakitStorageImpl::UpdateFormat1_3to1_4()
{
    int i, l, r, v, rank;

    /*
     * Get the markers view using the 1.3 format:
     */

    markers = storage->GetAs(E4_GRAPHMARKERS1_3);

    /*
     * Extract the root node.
     */

    r = (int) pFirst(unused[E4_GRAPHROOTNODE]);

    /*
     * For each marker, add a vertex to the root node with the marked
     * node as its value.
     */

    for (i = 0, l = markers.GetSize(); i < l; i++) {

	/*
	 * Only look at in use markers:
	 */

	if (((int) pFlags(markers[i]) & E4_INUSE) == 0) {
	    continue;
	}

	/*
	 * Add a new vertex to the root:
	 */

	v = AddVertex(r, E4_IOLAST, rank);
	if (v == E4_VERTEXNOTCREATED) {

	    /*
	     * Should abort.
	     */
	}

	/*
	 * Make the root node a parent of the marked node.
	 */

	AddParent(r, (int) pRoot(markers[i]), v);

	/*
	 * Now set its value to the node marked by the i'th marker:
	 */

	SetVertex(v, 
		  (int) pNameID(markers[i]),
		  E4_VTNODE,
		  pRoot(markers[i]));
    }

    /*
     * Discard the markers data.
     */

    markers.SetSize(0);

    /*
     * Now get the remaining views with the new format:
     */

    nodes = storage->GetAs(E4_GRAPHNODES1_4);
    vertices = storage->GetAs(E4_GRAPHVERTICES1_4);
    doubles = storage->GetAs(E4_GRAPHDOUBLES1_4);
    strings = storage->GetAs(E4_GRAPHSTRINGS1_4);
    binary = storage->GetAs(E4_GRAPHBINARY1_4);
    names = storage->GetAs(E4_GRAPHNAMES1_4);
    parents = storage->GetAs(E4_GRAPHNAMES1_4);

    pFirst(unused[E4_GRAPHSTORAGEMINORVER]) = E4_MKSTORAGE_MINORVER;

    /*
     * Commit the changed format.
     */

    storage->Commit();
}
		  
/*
 * Helper method to update the format from 1.2 to 1.3:
 */

void
e4_MetakitStorageImpl::UpdateFormat1_2to1_3()
{
    int i, l;

    /*
     * Get the nodes view using the 1.2 format:
     */

    nodes = storage->GetAs(E4_GRAPHNODES1_2);

    /*
     * Now add the "detachedvertices" property with the value
     * E4_NEXTNONE to each node.
     */

    for (i = 0, l = nodes.GetSize(); i < l; i++) {
	pDetachedVertices(nodes[i]) = E4_NEXTNONE;
    }

    /*
     * Now get the views with the new format:
     */

    markers = storage->GetAs(E4_GRAPHMARKERS1_3);
    nodes = storage->GetAs(E4_GRAPHNODES1_3);
    vertices = storage->GetAs(E4_GRAPHVERTICES1_3);
    doubles = storage->GetAs(E4_GRAPHDOUBLES1_3);
    strings = storage->GetAs(E4_GRAPHSTRINGS1_3);
    binary = storage->GetAs(E4_GRAPHBINARY1_3);
    names = storage->GetAs(E4_GRAPHNAMES1_3);
    parents = storage->GetAs(E4_GRAPHNAMES1_3);

    pFirst(unused[E4_GRAPHSTORAGEMINORVER]) = E4_MKSTORAGE_MINORVER;

    /*
     * Commit the changed format.
     */

    storage->Commit();
}

/*
 * Helper method to update the format from 1.1 to 1.2:
 */

void
e4_MetakitStorageImpl::UpdateFormat1_1to1_2()
{
    int i, l, pid, parentID, vv, pv;

    /*
     * Get all the views using the 1.1 format:
     */

    markers = storage->GetAs(E4_GRAPHMARKERS1_1);
    nodes = storage->GetAs(E4_GRAPHNODES1_1);
    vertices = storage->GetAs(E4_GRAPHVERTICES1_1);
    doubles = storage->GetAs(E4_GRAPHDOUBLES1_1);
    strings = storage->GetAs(E4_GRAPHSTRINGS1_1);
    binary = storage->GetAs(E4_GRAPHBINARY1_1);
    names = storage->GetAs(E4_GRAPHNAMES1_1);
    parents = storage->GetAs(E4_GRAPHPARENTS1_1);

    /*
     * Add the new properties:
     */

    vertices.AddProperty(pNextInParent);
    parents.AddProperty(pVertexChain);

    /*
     * For each node, traverse its parents and form a chain of
     * all vertices in that parent whose value is this node.
     */

    for (i = 0, l = nodes.GetSize(); i < l; i++) {
	/*
	 * If this index is not used, skip it.
	 */

	if (((int) pFlags(nodes[i]) & E4_INUSE) == 0) {
	    continue;
	}

	/*
	 * Found a node that is in use. Update its parent list so
	 * that each parent record will contain the list of vertices
	 * in that parent that point at this node.
	 */

	for (pid = (int) pParentID(nodes[i]);
	     pid != E4_NEXTNONE;
	     pid = (int) pNext(parents[pid])) {

	    /*
	     * Get the parent node.
	     */

	    parentID = (int) pNodeID(parents[pid]);

	    /*
	     * Iterate over its vertices, looking for vertices that
	     * have the current node as their value.
	     */

	    for (vv = (int) pFirstVertex(nodes[parentID]),
		   pv = E4_NEXTNONE;
		 vv != E4_NEXTNONE;
		 vv = (int) pNext(vertices[vv])) {

		/*
		 * See if this vertex has the current node as its value.
		 */

		if (((int) pVertexType(vertices[vv]) == E4_VTNODE) &&
		    ((int) pRowID(vertices[vv]) == i)) {

		    /*
		     * Put this vertex into the list of vertices
		     * reached from the parent record.
		     */

		    if (pv == E4_NEXTNONE) {
			pVertexChain(parents[pid]) = vv;
		    } else {
			pNextInParent(vertices[pv]) = vv;
		    }

		    /*
		     * Remember previous link in chain.
		     */

		    pv = vv;
		}
	    }
	}
    }

    /*
     * Now get the views with the new format:
     */

    markers = storage->GetAs(E4_GRAPHMARKERS1_2);
    nodes = storage->GetAs(E4_GRAPHNODES1_2);
    vertices = storage->GetAs(E4_GRAPHVERTICES1_2);
    doubles = storage->GetAs(E4_GRAPHDOUBLES1_2);
    strings = storage->GetAs(E4_GRAPHSTRINGS1_2);
    binary = storage->GetAs(E4_GRAPHBINARY1_2);
    names = storage->GetAs(E4_GRAPHNAMES1_2);
    parents = storage->GetAs(E4_GRAPHNAMES1_2);

    pFirst(unused[E4_GRAPHSTORAGEMINORVER]) = E4_MKSTORAGE_MINORVER;

    storage->Commit();
}

/*
 * Helper method to update the format from 1.0 to 1.1:
 */

void
e4_MetakitStorageImpl::UpdateFormat1_0to1_1()
{
    int i, l;

    /*
     * Get all the views using the 1.0 format:
     */

    markers = storage->GetAs(E4_GRAPHMARKERS1_0);
    nodes = storage->GetAs(E4_GRAPHNODES1_0);
    vertices = storage->GetAs(E4_GRAPHVERTICES1_0);
    doubles = storage->GetAs(E4_GRAPHDOUBLES1_0);
    strings = storage->GetAs(E4_GRAPHSTRINGS1_0);
    binary = storage->GetAs(E4_GRAPHBINARY1_0);
    names = storage->GetAs(E4_GRAPHNAMES1_0);
    parents = storage->GetAs(E4_GRAPHPARENTS1_0);

    /*
     * Add the new property "flags":
     */

    markers.AddProperty(pFlags);
    nodes.AddProperty(pFlags);
    vertices.AddProperty(pFlags);
    doubles.AddProperty(pFlags);
    strings.AddProperty(pFlags);
    binary.AddProperty(pFlags);
    names.AddProperty(pFlags);
    parents.AddProperty(pFlags);

    /*
     * For each entry, check if its "used" propert was set -- if so, set
     * the "flags" property to E4_INUSE.
     */

    for (i = 0, l = markers.GetSize(); i < l; i++) {
	if ((int) pUsed(markers[i]) == 1) {
	    pFlags(markers[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = nodes.GetSize(); i < l; i++) {
	if ((int) pUsed(nodes[i]) == 1) {
	    pFlags(nodes[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = vertices.GetSize(); i < l; i++) {
	if ((int) pUsed(vertices[i]) == 1) {
	    pFlags(vertices[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = doubles.GetSize(); i < l; i++) {
	if ((int) pUsed(doubles[i]) == 1) {
	    pFlags(doubles[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = strings.GetSize(); i < l; i++) {
	if ((int) pUsed(strings[i]) == 1) {
	    pFlags(strings[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = binary.GetSize(); i < l; i++) {
	if ((int) pUsed(binary[i]) == 1) {
	    pFlags(binary[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = names.GetSize(); i < l; i++) {
	if ((int) pUsed(names[i]) == 1) {
	    pFlags(names[i]) = (int) E4_INUSE;
	}
    }
    for (i = 0, l = parents.GetSize(); i < l; i++) {
	if ((int) pUsed(parents[i]) == 1) {
	    pFlags(parents[i]) = (int) E4_INUSE;
	}
    }

    /*
     * Now discard the old "used" property by getting the views with
     * the new format:
     */

    markers = storage->GetAs(E4_GRAPHMARKERS1_1);
    nodes = storage->GetAs(E4_GRAPHNODES1_1);
    vertices = storage->GetAs(E4_GRAPHVERTICES1_1);
    doubles = storage->GetAs(E4_GRAPHDOUBLES1_1);
    strings = storage->GetAs(E4_GRAPHSTRINGS1_1);
    binary = storage->GetAs(E4_GRAPHBINARY1_1);
    names = storage->GetAs(E4_GRAPHNAMES1_1);
    parents = storage->GetAs(E4_GRAPHNAMES1_1);

    pFirst(unused[E4_GRAPHSTORAGEMINORVER]) = E4_MKSTORAGE_MINORVER;

    storage->Commit();
}

/*
 * The destructor:
 */

e4_MetakitStorageImpl::~e4_MetakitStorageImpl()
{
    /*
     * If the Metakit storage is still open, close it.
     */

    if (storage != NULL) {
        delete storage;
    }

    /*
     * Clean up the deletion stacks if they're present.
     */

    if (idStack1 != NULL) {
	delete idStack1;
    }
    if (idStack2 != NULL) {
	delete idStack2;
    }
}
        
/*
 * Commit changes to this storage.
 */

bool
e4_MetakitStorageImpl::Commit()
{
    if (!IsDirty()) {
	return true;
    }
    storage->Commit();
    MarkClean();

    return true;
}

/*
 * Helper functions to initialize and manage the free lists.
 */

void
e4_MetakitStorageImpl::PopulateNameHash()
{
    int i, lim;

    for (i = 0, lim = names.GetSize(); i < lim; i++) {
	if (((int) pFlags(names[i]) & E4_INUSE) == E4_INUSE) {
            AddNameToNameHash(pNameVal(names[i]), i);
	}
    }
}

void
e4_MetakitStorageImpl::MakeParentSpace()
{
    int next = parents.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    parents.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) = next;
    for (; next < newsize; next++) {
	pNext(parents[next]) = next + 1;
	pFlags(parents[next]) = 0;
    }
    pNext(parents[newsize - 1]) = E4_NEXTNONE;
    pFlags(parents[newsize - 1]) = 0;
}

void
e4_MetakitStorageImpl::MakeNodeSpace()
{
    int next = nodes.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    nodes.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) = next;
    for (; next < newsize; next++) {
        pNext(nodes[next]) = next + 1;
	pFlags(nodes[next]) = 0;
    }
    pNext(nodes[newsize - 1]) = E4_NEXTNONE;
    pFlags(nodes[newsize - 1]) = 0;

    statistics[(int) E4_SPNODE][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeVertexSpace()
{
    int next = vertices.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    vertices.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) = next;
    for (; next < newsize; next++) {
        pNext(vertices[next]) = next + 1;
	pFlags(vertices[next]) = 0;
    }
    pNext(vertices[newsize - 1]) = E4_NEXTNONE;
    pFlags(vertices[newsize - 1]) = 0;

    statistics[(int) E4_SPVERTEX][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeDoubleSpace()
{
    int next = doubles.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    doubles.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]) = next;
    for (; next < newsize; next++) {
        pNext(doubles[next]) = next + 1;
	pFlags(doubles[next]) = 0;
    }
    pNext(doubles[newsize - 1]) = E4_NEXTNONE;
    pFlags(doubles[newsize - 1]) = 0;

    statistics[(int) E4_SPDOUBLE][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeStringSpace()
{
    int next = strings.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    strings.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]) = next;
    for (; next < newsize; next++) {
        pNext(strings[next]) = next + 1;
	pFlags(strings[next]) = 0;
    }
    pNext(strings[newsize - 1]) = E4_NEXTNONE;
    pFlags(strings[newsize - 1]) = 0;

    statistics[(int) E4_SPSTRING][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeBinarySpace()
{
    int next = binary.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    binary.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]) = next;
    for (; next < newsize; next++) {
        pNext(binary[next]) = next + 1;
	pFlags(binary[next]) = 0;
    }
    pNext(binary[newsize - 1]) = E4_NEXTNONE;
    pFlags(binary[newsize - 1]) = 0;

    statistics[(int) E4_SPBINARY][(int) E4_SSAVAIL] = newsize;
}

void
e4_MetakitStorageImpl::MakeNameSpace()
{
    int next = names.GetSize();
    int newsize = next + E4_GRAPHINCREMENT;

    names.SetSize(newsize);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]) = next;
    for (; next < newsize; next++) {
	pNext(names[next]) = next + 1;
	pFlags(names[next]) = 0;
    }
    pNext(names[newsize - 1]) = E4_NEXTNONE;
    pFlags(names[newsize - 1]) = 0;

    statistics[(int) E4_SPNAME][(int) E4_SSAVAIL] = newsize;
}

/*
 * Compute the rank of a child node in the vertex list of a node.
 */

int
e4_MetakitStorageImpl::GetRankOfChildNode(int parentID, int childID) const
{
    int f, rank;
    
    if ((parentID < 0) || (parentID >= nodes.GetSize()) ||
        (((int) pFlags(nodes[parentID]) & E4_INUSE) == 0) ||
        (childID < 0) || (childID >= nodes.GetSize()) ||
        (((int) pFlags(nodes[childID]) & E4_INUSE) == 0)) {
        return E4_VERTEXNOTFOUND;
    }
    for (rank = 1, f = (int) pFirstVertex(nodes[parentID]);
         f != E4_NEXTNONE;
         f = (int) pNext(vertices[f]), rank++) {
        if (((e4_VertexType) (int) pVertexType(vertices[f]) == E4_VTNODE) &&
            ((int) pRowID(vertices[f]) == childID)) {
            return rank;
        }
    }
    return E4_VERTEXNOTFOUND;
}

/*
 * Helper function to set a vertex value:
 */

bool
e4_MetakitStorageImpl::SetVertex(int index, int nameID, 
				 int vertexType, int itemID)
{
    pVertexType(vertices[index]) = vertexType;
    pNameID(vertices[index]) = nameID;
    pRowID(vertices[index]) = itemID;

    /*
     * Do not change the "next" and "used" property.
     */

    return true;
}

/*
 * Reserve a vertex for a detached vertex:
 */

int
e4_MetakitStorageImpl::ReserveVertexID(const char *nm)
{
    int i;

    /*
     * First of all grab a free row in the "vertices" view.
     */

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) == E4_NEXTNONE) {
	MakeVertexSpace();
    }
    i = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) = (int) pNext(vertices[i]);

    pNodeID(vertices[i]) = E4_NEXTNONE;
    pVertexType(vertices[i]) = E4_VTINT;
    pRowID(vertices[i]) = 0;

    /*
     * The vertex starts out detached. Since we do not want to receive
     * a detach callback on this vertex (if it stays detached) we mark
     * it as already having had its callback.
     */

    pFlags(vertices[i]) = (E4_INUSE | E4_DETACHED | E4_DETACHNOTIFY);
    if (nm != NULL) {
	pNameID(vertices[i]) = InternName(nm);
    }

    return i;
}

/*
 * Insert a new vertex in the node identified by nodeID at the position
 * indicated by "io" and "rank". The ID of the new vertex is returned.
 */

int
e4_MetakitStorageImpl::AddVertex(int nodeID, e4_InsertOrder io, int &rank)
{
    int i, j, k, flags;

    i = ReserveVertexID(NULL);

    /*
     * Update the count of used vertices.
     */

    statistics[(int) E4_SPVERTEX][(int) E4_SSUSED]++;
    statistics[(int) E4_SPVERTEX][(int) E4_SSALLOC]++;

    /*
     * Initialize the values in the newly allocated vertex.
     */

    pNameID(vertices[i]) = -1;
    pNodeID(vertices[i]) = nodeID;
    pRowID(vertices[i]) = -1;
    pUserData(vertices[i]) = 0;

    /*
     * The vertex is no longer detached.
     */

    flags = (int) pFlags(vertices[i]);
    flags &= ~(E4_DETACHED | E4_DETACHNOTIFY);
    pFlags(vertices[i]) = flags;

    /*
     * Now use the row we grabbed as the vertex indicated by the "io" and
     * "rank" arguments.
     */

    switch (io) {
    case E4_IOFIRST:
	pNext(vertices[i]) = (int) pFirstVertex(nodes[nodeID]);
	pPrev(vertices[i]) = E4_NEXTNONE;
	if ((int) pFirstVertex(nodes[nodeID]) != E4_NEXTNONE) {
	    pPrev(vertices[pFirstVertex(nodes[nodeID])]) = i;
	}
        pFirstVertex(nodes[nodeID]) = i;
        if ((int) pLastVertex(nodes[nodeID]) == E4_NEXTNONE) {
            pLastVertex(nodes[nodeID]) = i;
	}
	rank = 1;
	break;

    case E4_IOLAST:
	pNext(vertices[i]) = E4_NEXTNONE;
        if ((int) pFirstVertex(nodes[nodeID]) == E4_NEXTNONE) {
            pFirstVertex(nodes[nodeID]) = i;
        }
        if ((int) pLastVertex(nodes[nodeID]) != E4_NEXTNONE) {
	    pPrev(vertices[i]) = (int) pLastVertex(nodes[nodeID]);
            pNext(vertices[pLastVertex(nodes[nodeID])]) = i;
	} else {
	    pPrev(vertices[i]) = E4_NEXTNONE;
	}
        pLastVertex(nodes[nodeID]) = i;
        rank = (int) pVertexCount(nodes[nodeID]) + 1;
	break;

    case E4_IOAT:
    case E4_IOBEFORE:
	if (rank == 1) {
	    pNext(vertices[i]) = (int) pFirstVertex(nodes[nodeID]);
	    pPrev(vertices[i]) = E4_NEXTNONE;
            if ((int) pLastVertex(nodes[nodeID]) == E4_NEXTNONE) {
                pLastVertex(nodes[nodeID]) = i;
	    }
	    if ((int) pFirstVertex(nodes[nodeID]) != E4_NEXTNONE) {
		pPrev(vertices[pFirstVertex(nodes[nodeID])]) = i;
	    }
	    pFirstVertex(nodes[nodeID]) = i;
	    rank = 1;
	} else {
	    j = VertexIDFromRank(nodeID, rank - 1);
	    if (j == E4_VERTEXNOTFOUND) {
                goto error;
            }
            k = (int) pNext(vertices[j]);
	    pNext(vertices[j]) = i;
	    pNext(vertices[i]) = k;
	    pPrev(vertices[k]) = i;
	    pPrev(vertices[i]) = j;
            if ((int) pLastVertex(nodes[nodeID]) == j) {
                pLastVertex(nodes[nodeID]) = i;
            }
	    rank = i;
	}
	break;

    case E4_IOAFTER:
        j = VertexIDFromRank(nodeID, rank);
	if (j == E4_VERTEXNOTFOUND) {
	    goto error;
	}
	pNext(vertices[i]) = (int) pNext(vertices[j]);
	pPrev(vertices[i]) = j;
	if ((int) pNext(vertices[i]) != E4_NEXTNONE) {
	    pPrev(vertices[pNext(vertices[i])]) = i;
	}
	pNext(vertices[j]) = i;
        if ((int) pLastVertex(nodes[nodeID]) == j) {
            pLastVertex(nodes[nodeID]) = i;
        }
	rank = j + 1;
	break;
    }

    /*
     * Update the vertex count for this node.
     */

    pVertexCount(nodes[nodeID]) = (int) pVertexCount(nodes[nodeID]) + 1;

    return i;

error:
    /*
     * Some kind of error occurred, after we allocated the new row.
     * Free up the allocated row so that it can be used in a future
     * operation.
     */

    AddToUnusedVerticesList(i);

    return E4_VERTEXNOTCREATED;
}

/*
 * Free a row occupied by a 64-bit floating point value.
 */

bool
e4_MetakitStorageImpl::FreeDouble(int doubleID)
{
    if ((doubleID < 0) || 
	(doubleID >= doubles.GetSize()) ||
        (((int) pFlags(doubles[doubleID]) & E4_INUSE) == 0)) {
        return false;
    }

    AddToUnusedDoublesList(doubleID);
    MarkDirty();

    return true;
}

/*
 * Free a row occupied by a string value. Release the storage occupied by
 * the actual string value also.
 */

bool
e4_MetakitStorageImpl::FreeString(int stringID)
{
    if ((stringID < 0) || 
	(stringID >= strings.GetSize()) ||
        (((int) pFlags(strings[stringID]) & E4_INUSE) == 0)) {
        return false;
    }

    AddToUnusedStringsList(stringID);
    pStringVal(strings[stringID]) = "";
    MarkDirty();

    return true;
}

/*
 * Free a row occupied by a binary value. Release the storage occupied by
 * the actual value also.
 */

bool
e4_MetakitStorageImpl::FreeBinary(int binaryID)
{
    if ((binaryID < 0) || 
	(binaryID >= binary.GetSize()) ||
        (((int) pFlags(binary[binaryID]) & E4_INUSE) == 0)) {
        return false;
    }

    AddToUnusedBinaryList(binaryID);

    c4_Bytes bytes(NULL, 0);

    pBinaryVal(binary[binaryID]) = bytes;
    MarkDirty();

    return true;
}

/*
 * Return a node from a designated row.
 */

bool
e4_MetakitStorageImpl::GetNode(int index, e4_NodeImpl *&value) const
{
    if ((index < 0) ||
	(index >= nodes.GetSize()) ||
        (((int) pFlags(nodes[index]) & E4_INUSE) == 0)) {
        return false;
    }

    value = FindOrCreateNode(index);
    return true;
}

/*
 * Specialized internal version to return a node from a designated row.
 */

bool
e4_MetakitStorageImpl::GetNode(int index, e4_NodeImpl *&value, 
			       bool *isNew) const
{
    if ((index < 0) ||
	(index >= nodes.GetSize()) ||
        (((int) pFlags(nodes[index]) & E4_INUSE) == 0)) {
        return false;
    }

    value = FindOrCreateNode(index, isNew);
    return true;
}

/*
 * Return an integer value from a designated row.
 */

bool
e4_MetakitStorageImpl::GetInt(int index, int &value) const 
{
    /*
     * Integers are represented directly in the vertex, so the index
     * *is* the vertex value.
     */

    value = index;

    return true;
}

/*
 * Return a 64-bit floating point value from a designated row.
 */

bool
e4_MetakitStorageImpl::GetDouble(int index, double &value) const
{
    if ((index < 0) ||
	(index >= doubles.GetSize()) ||
        (((int) pFlags(doubles[index]) & E4_INUSE) == 0)) {
        return false;
    }

    value = (double) pDoubleVal(doubles[index]);

    return true;
}

/*
 * Return a string value from a designated row.
 */

bool
e4_MetakitStorageImpl::GetString(int index, char *&value) const
{
    if ((index < 0) ||
	(index >= strings.GetSize()) ||
        (((int) pFlags(strings[index]) & E4_INUSE) == 0)) {
        return false;
    }

    value = (char *) (const char *) pStringVal(strings[index]);

    return true;
}

/*
 * Return a binary value from a designated row.
 */

bool
e4_MetakitStorageImpl::GetBinary(int index, void *&value,
				 int &nbytes) const
{
    if ((index < 0) ||
	(index >= binary.GetSize()) ||
        (((int) pFlags(binary[index]) & E4_INUSE) == 0)) {
        return false;
    }

    c4_Bytes bytes = pBinaryVal(binary[index]);

    value = (void *) (const void *) bytes.Contents();
    nbytes = bytes.Size();

    return true;
}

/*
 * Retrieve a vertex name by the name ID.
 */

bool
e4_MetakitStorageImpl::GetVertexName(int index, const char *&nm) const
{
    if ((index < 0) ||
	(index >= names.GetSize()) ||
	(((int) pFlags(names[index]) & E4_INUSE) == 0)) {
	return false;
    }

    nm = pNameVal(names[index]);
    return true;
}

/*
 * Returns true if the given ID is that of the root node.
 */

bool
e4_MetakitStorageImpl::IsRootNodeID(int nodeID)
{
    if ((int) pFirst(unused[E4_GRAPHROOTNODE]) == nodeID) {
	return true;
    }
    return false;
}

/*
 * Get the ID of the root node.
 */

int
e4_MetakitStorageImpl::GetRootNodeID()
{
    return (int) pFirst(unused[E4_GRAPHROOTNODE]);
}

/*
 * Set the root node to another node.
 */

bool
e4_MetakitStorageImpl::SetRootNodeID(int newRootID)
{
    /*
     * Sanity checks.
     */

    if ((newRootID < 0) || (newRootID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[newRootID]) & E4_INUSE) != E4_INUSE)) {
	return false;
    }

    /*
     * Set the new root ID:
     */

    pFirst(unused[E4_GRAPHROOTNODE]) = newRootID;

    return true;
}

/*
 * Add a row for a new node.
 */

int
e4_MetakitStorageImpl::ReserveNodeID()
{
    int idx;

    if (pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) == E4_NEXTNONE) {
        MakeNodeSpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) = (int) pNext(nodes[idx]);

    statistics[(int) E4_SPNODE][(int) E4_SSUSED]++;
    statistics[(int) E4_SPNODE][(int) E4_SSALLOC]++;

    /*
     * The node starts out as detached. We don't want to receive a detach
     * callback on this node (if it stays detached) so we mark it as having
     * already had a callback.
     */

    pFlags(nodes[idx]) = (E4_INUSE | E4_DETACHED | E4_DETACHNOTIFY);
    pUserData(nodes[idx]) = 0;
    pNext(nodes[idx]) = E4_NEXTNONE;
    pVertexCount(nodes[idx]) = 0;
    pFirstVertex(nodes[idx]) = E4_NEXTNONE;
    pLastVertex(nodes[idx]) = E4_NEXTNONE;
    pParentID(nodes[idx]) = E4_NEXTNONE;
    pRefCount(nodes[idx]) = 0;
    pDetachedVertices(nodes[idx]) = E4_NEXTNONE;

    MarkDirty();
    return idx;
}

/*
 * Add an integer to the storage -- nothing to do because the
 * integer is represented directly in the vertex.
 */

int
e4_MetakitStorageImpl::AddInt(int value)
{
    MarkDirty();
    return value;
}

/*
 * Add a row for a new 64-bit floating point value.
 */

int
e4_MetakitStorageImpl::AddDouble(double value)
{
    int idx;

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]) == E4_NEXTNONE) {
        MakeDoubleSpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]) = (int) pNext(doubles[idx]);

    statistics[(int) E4_SPDOUBLE][(int) E4_SSUSED]++;
    statistics[(int) E4_SPDOUBLE][(int) E4_SSALLOC]++;

    pFlags(doubles[idx]) = E4_INUSE;
    pDoubleVal(doubles[idx]) = value;
    MarkDirty();
    return idx;
}

/*
 * Add a row for a new string value.
 */

int
e4_MetakitStorageImpl::AddString(const char *value)
{
    int idx;

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]) == E4_NEXTNONE) {
        MakeStringSpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]) = (int) pNext(strings[idx]);

    statistics[(int) E4_SPSTRING][(int) E4_SSUSED]++;
    statistics[(int) E4_SPSTRING][(int) E4_SSALLOC]++;

    pFlags(strings[idx]) = E4_INUSE;
    pStringVal(strings[idx]) = value;
    MarkDirty();
    return idx;
}

/*
 * Add a row for a new binary value.
 */

int
e4_MetakitStorageImpl::AddBinary(const void *value, int nbytes)
{
    c4_Bytes bytes(value, nbytes);
    int idx;

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]) == E4_NEXTNONE) {
        MakeBinarySpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]) = (int) pNext(binary[idx]);

    statistics[(int) E4_SPBINARY][(int) E4_SSUSED]++;
    statistics[(int) E4_SPBINARY][(int) E4_SSALLOC]++;

    pFlags(binary[idx]) = E4_INUSE;
    pBinaryVal(binary[idx]) = bytes;
    MarkDirty();
    return idx;
}

/*
 * Add a row for a new vertex, node or marker name.
 */

int
e4_MetakitStorageImpl::AddName(const char *nm)
{
    int idx;

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]) == E4_NEXTNONE) {
	MakeNameSpace();
    }
    idx = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]) = (int) pNext(names[idx]);

    statistics[(int) E4_SPNAME][(int) E4_SSUSED]++;
    statistics[(int) E4_SPNAME][(int) E4_SSALLOC]++;

    pFlags(names[idx]) = E4_INUSE;
    pNameVal(names[idx]) = nm;
    MarkDirty();
    return idx;
}

/*
 * Is this instance valid?
 */

bool
e4_MetakitStorageImpl::IsValid() const
{
    return (storage == NULL) ? false : true;
}

int
e4_MetakitStorageImpl::VertexIDFromNthVertex(int nodeID, int nameID,
					     int nth, int &rank) const
{
    int i, r, cnt;

    /*
     * Its not cached, so we must search for it the slow way. Note
     * that it's also not known whether there is an nth vertex with
     * name nameID in this node.
     */

    for (r = 1, i = (int) pFirstVertex(nodes[nodeID]), cnt = 0; 
	 i != E4_NEXTNONE; 
	 i = (int) pNext(vertices[i]), r++) {
        if ((int) pNameID(vertices[i]) == nameID) {
            cnt++;
            if (cnt == nth) {
                break;
            }
        }
    }
    if ((cnt != nth) || (i == E4_NEXTNONE) || 
	((int) pNameID(vertices[i]) != nameID)) {
        return E4_VERTEXNOTFOUND;
    }
    rank = r;
    return i;
}

int
e4_MetakitStorageImpl::VertexIDFromRank(int nodeID, int rank) const
{
    int i, r;

    /*
     * Quick check to see if the node has at least that many
     * vertices. If not, we do not need to search.
     */

    if ((int) pVertexCount(nodes[nodeID]) < rank) {
	return E4_VERTEXNOTFOUND;
    }

    /*
     * The data we were looking for is not cached, so we
     * must search for it the slow way. But we know at this
     * point that the search will be successful.
     */

    for (i = (int) pFirstVertex(nodes[nodeID]), r = 1; 
	 r < rank;
         i = (int) pNext(vertices[i]), r++) {
    }
    return i;
}

int
e4_MetakitStorageImpl::RankFromVertexID(int nodeID, int index) const
{
    int i, r;

    for (i = (int) pFirstVertex(nodes[nodeID]), r = 1;
         (i != E4_NEXTNONE) && (i != index);
         i = (int) pNext(vertices[i]), r++) {
    }
    if (i != index) {
        r = E4_VERTEXNOTFOUND;
    }
    return r;
}

int
e4_MetakitStorageImpl::GetVertexIDInParent(int parentID, int childID) const 
{
    int i;

    /*
     * Sanity checks.
     */

    if ((parentID < 0) ||
	(parentID >= nodes.GetSize()) ||
        (((int) pFlags(nodes[parentID]) & E4_INUSE) == 0) ||
        (childID < 0) ||
	(childID >= nodes.GetSize()) ||
        (((int) pFlags(nodes[childID]) & E4_INUSE) == 0)) {
        return E4_VERTEXNOTFOUND;
    }

    /*
     * Look for first vertex in parent that contains
     * the child node as its value. We cannot use the
     * nextInParent chain because that's not guaranteed
     * to be in rank order.
     */

    for (i = (int) pFirstVertex(nodes[parentID]); 
	 i != E4_NEXTNONE;
	 i = (int) pNext(vertices[i])) {
        if (((e4_VertexType) (int) pVertexType(vertices[i]) == E4_VTNODE) &&
            ((int) pRowID(vertices[i]) == childID)) {
            return i;
        }
    }
    return E4_VERTEXNOTFOUND;
}

/*
 * Free the storage occupied by an old value of a vertex:
 */

void
e4_MetakitStorageImpl::FreeVertexValue(int vertexID)
{
    int x = 0;
    switch (pVertexType(vertices[vertexID])) {
    case E4_VTNODE:
	RemoveParent((int) pRowID(vertices[vertexID]),
		     (int) pNodeID(vertices[vertexID]), 
		     vertexID,
		     true);
	break;
    case E4_VTINT:

        /*
	 * Nothing to do because the value is represented directly
	 * in the vertex.
	 */

        break;
    case E4_VTDOUBLE:
        FreeDouble((int) pRowID(vertices[vertexID]));
        break;
    case E4_VTSTRING:
        FreeString((int) pRowID(vertices[vertexID]));
        break;
    case E4_VTBINARY:
        FreeBinary((int) pRowID(vertices[vertexID]));
        break;
    default:
        /*
         * Should abort here.
         */

        break;
    }
}

/*
 * Get the value of a vertex given its index.
 */

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, e4_ValueImpl *&v) const
{
    e4_ValueImpl *vp;
    e4_NodeImpl *n;
    double d;
    char *s;
    void *bytes;
    int nbytes;

    vp = new e4_ValueImpl;
    vp->vertexType = (e4_VertexType) (int)  pVertexType(vertices[i]);
    switch(vp->vertexType) {
    case E4_VTUNKNOWN:
	return false;
    case E4_VTNODE:
        if (GetNode((int) pRowID(vertices[i]), n) != true) {
	    return false;
	}
	vp->u.n = n;
	break;
    case E4_VTINT:

        /*
	 * Integers are represented directly in a vertex.
	 */

	vp->u.i = (int) pRowID(vertices[i]);
	break;
    case E4_VTDOUBLE:
        if (GetDouble((int) pRowID(vertices[i]), d) != true) {
	    return false;
	}
	vp->u.d = d;
	break;
    case E4_VTSTRING:
	if (GetString((int) pRowID(vertices[i]), s) != true) {
	    return false;
	}
	vp->u.s = s;
	break;
    case E4_VTBINARY:
	if (GetBinary((int) pRowID(vertices[i]), bytes, nbytes) != true) {
	    return false;
	}
	vp->u.b.bytes = bytes;
	vp->u.b.nbytes = nbytes;
	break;
    default:
	return false;
    }

    v = vp;
    return true;
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, e4_NodeImpl *&n) const
{
    if ((int) pVertexType(vertices[i]) != E4_VTNODE) {
	return false;
    }
    return GetNode((int) pRowID(vertices[i]), n);
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, int &value) const
{
    if ((int) pVertexType(vertices[i]) != E4_VTINT) {
	return false;
    }

    /*
     * Integers are represented directly in the vertex.
     */

    value = (int) pRowID(vertices[i]);
    return true;
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, double &value) const
{
    if ((int) pVertexType(vertices[i]) != E4_VTDOUBLE) {
	return false;
    }
    return GetDouble((int) pRowID(vertices[i]), value);
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, const char *&value) const
{
    char *lvalue;

    if ((int) pVertexType(vertices[i]) != E4_VTSTRING) {
	return false;
    }
    if (GetString((int) pRowID(vertices[i]), lvalue)) {
	value = lvalue;
	return true;
    }
    return false;
}

bool
e4_MetakitStorageImpl::GetVertexByIndex(int i, const void *&bytes,
                                   int &nbytes) const
{
    void *lbytes;

    if ((int) pVertexType(vertices[i]) != E4_VTBINARY) {
	return false;
    }
    if (GetBinary((int) pRowID(vertices[i]), lbytes, nbytes)) {
	bytes = lbytes;
        return true;
    }
    return false;
}

/*
 * Set a vertex whose index is given:
 */

bool
e4_MetakitStorageImpl::SetVertexByIndex(int i, int value)
{
    if (pVertexType(vertices[i]) != E4_VTINT) {
        FreeVertexValue(i);
    }
    SetVertex(i, pNameID(vertices[i]), E4_VTINT, value);
    MarkDirty();
    return true;
}

bool
e4_MetakitStorageImpl::SetVertexByIndex(int i, double value)
{
    int j;

    if ((int) pVertexType(vertices[i]) != E4_VTDOUBLE) {
	FreeVertexValue(i);
	j = AddDouble(value);
	SetVertex(i, (int) pNameID(vertices[i]), E4_VTDOUBLE, j);
    } else {
	pDoubleVal(doubles[pRowID(vertices[i])]) = value;
    }
    MarkDirty();
    return true;
}

bool
e4_MetakitStorageImpl::SetVertexByIndex(int i, const char *value)
{
    int j;

    if ((int) pVertexType(vertices[i]) != E4_VTSTRING) {
	FreeVertexValue(i);
	j = AddString(value);
	SetVertex(i, (int) pNameID(vertices[i]), E4_VTSTRING, j);
    } else {

        /*
	 * NOTE: This assumes that string values are not
	 * shared.
	 */

	pStringVal(strings[pRowID(vertices[i])]) = value;
    }
    MarkDirty();
    return true;
}

bool
e4_MetakitStorageImpl::SetVertexByIndex(int i, const void *bs, int nbs)
{
    int j;

    if ((int) pVertexType(vertices[i]) != E4_VTBINARY) {
	FreeVertexValue(i);
	j = AddBinary(bs, nbs);
	SetVertex(i, pNameID(vertices[i]), E4_VTBINARY, j);
    } else {
        c4_Bytes bytes(bs, nbs);
	pBinaryVal(binary[pRowID(vertices[i])]) = bytes;
    }
    MarkDirty();
    return true;
}

bool
e4_MetakitStorageImpl::SetVertexByIndexToNode(int i, int childID)
{
    MarkDirty();
    FreeVertexValue(i);
    SetVertex(i, pNameID(vertices[i]), E4_VTNODE, childID);
    AddParent(childID,
	      (int) pNodeID(vertices[i]),
	      i);

    return true;
}

bool
e4_MetakitStorageImpl::SetNodeUserData(int nodeID, int userData) const
{
    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return false;
    }

    pUserData(nodes[nodeID]) = userData;

    return true;
}

bool
e4_MetakitStorageImpl::SetVertexUserData(int vertexID, int userData) const
{
    if ((vertexID < 0) ||
	(vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return false;
    }

    pUserData(vertices[vertexID]) = userData;

    return true;
}

bool
e4_MetakitStorageImpl::GetNodeUserData(int nodeID, int &userData) const
{
    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return false;
    }

    userData = (int) pUserData(nodes[nodeID]);

    return true;
}

bool
e4_MetakitStorageImpl::GetVertexUserData(int vertexID, int &userData) const
{
    if ((vertexID < 0) || 
	(vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return false;
    }

    userData = (int) pUserData(vertices[vertexID]);

    return true;
}

/*
 * Given a node ID, retrieve the parent node or its ID.
 */

e4_NodeImpl *
e4_MetakitStorageImpl::GetParentNode(int childID, int nth) const 
{
    e4_NodeImpl *n;
    int parentID;

    if ((childID < 0) ||
	(childID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[childID]) & E4_INUSE) == 0)) {
	return NULL;
    }
    parentID = GetParentNodeID(childID, nth);
    if (parentID == E4_NODENOTFOUND) {
	return NULL;
    }
    if (GetNode(parentID, n) == false) {
	return NULL;
    }
    return n;
}

int 
e4_MetakitStorageImpl::GetParentNodeID(int childID, int nth) const 
{
    int i, pid;

    /*
     * If this node has no parents then return an indication
     * that we could not find an ID for its parent.
     */

    if ((int) pParentID(nodes[childID]) == E4_NEXTNONE) {
	return E4_NODENOTFOUND;
    }

    /*
     * If nth is less than one, we want the last parent, whatever its rank.
     */

    if (nth < 1) {
	for (pid = (int) pParentID(nodes[childID]);
	     (int) pNext(parents[pid]) != E4_NEXTNONE;
	     pid = (int) pNext(parents[pid]));
	return (int) pNodeID(parents[pid]);
    }

    /*
     * Get the nth parent in the list of parents.
     */

    for (pid = (int) pParentID(nodes[childID]), i = 1; 
	 (i < nth) && ((int) pNext(parents[pid]) != E4_NEXTNONE);
	 i++, pid = (int) pNext(parents[pid]));

    /*
     * If this node does not have an nth parent, return failure.
     */

    if (i < nth) {
	return E4_NODENOTFOUND;
    }

    /*
     * We found the nth parent.
     */

    return (int) pNodeID(parents[pid]);
}

/*
 * Get the vertex after one identified by vertexID.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::NextVertexAfter(int num, int vertexID) const
{
    int nvID;
    int i;

    if ((vertexID < 0) ||
	(vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return NULL;
    }
    if (num < 1) {
	return NULL;
    }
    for (nvID = vertexID, i = 0; i < num; i++) {
	nvID = (int) pNext(vertices[nvID]);
	if ((nvID == E4_NEXTNONE) ||
	    (((int) pFlags(vertices[nvID]) & E4_INUSE) == 0)) {
	    return NULL;
	}
    }
    return GetVertex(nvID);
}

/*
 * Get the vertex before one identified by vertexID.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::PrevVertexBefore(int num, int vertexID) const
{
    int nvID;
    int i;

    if ((vertexID < 0) ||
	(vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return NULL;
    }
    if (num < 1) {
	return NULL;
    }
    for (nvID = vertexID, i = 0; i < num; i++) {
	nvID = (int) pPrev(vertices[nvID]);
	if ((nvID == E4_NEXTNONE) || 
	    (((int) pFlags(vertices[nvID]) & E4_INUSE) == 0)) {
	    return NULL;
	}
    }
    return GetVertex(nvID);
}

/*
 * Is the node identified by index valid?
 */

bool
e4_MetakitStorageImpl::IsLegalNodeID(int index) const
{
    int flags;

    /*
     * If index is out of bounds, return false.
     */

    if ((index < 0) || (index >= nodes.GetSize())) {
	return false;
    }

    /*
     * If not E4_INUSE, return false.
     */

    flags = (int) pFlags(nodes[index]);
    if ((flags & E4_INUSE) == 0) {
	return false;
    }

    /*
     * If not detached, return true.
     */

    if ((flags & E4_DETACHED) == 0) {
	return true;
    }

    /*
     * If we get here, the node is detached.
     *
     * The node must either be the root node or referenced by the
     * user program, to be legal.
     */

    if (!((((e4_MetakitStorageImpl *) this)->IsRootNodeID(index)) ||
	  IsReferencedNode(index))) {
	return false;
    }
    return true;
}

/*
 * Retrieve the ID of the first and last vertex in a node.
 */

int
e4_MetakitStorageImpl::GetFirstVertexID(int nodeID) const 
{
    return (int) pFirstVertex(nodes[nodeID]);
}

int 
e4_MetakitStorageImpl::GetLastVertexID(int nodeID) const 
{
    return (int) pLastVertex(nodes[nodeID]);
}

/*
 * Retrieve an e4_VertexImpl instance for the vertex following the one
 * identified by the given ID in its node.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertex(int vertexID, e4_VisitMethod vm,
				      int vf, int nameID, int nodeID,
				      e4_VertexType typeID) const
{
    switch (vm) {
    case E4_VMSTORAGE:
        return FindNextVertexStorage(vertexID, vf, nameID, typeID);
    case E4_VMNODE:
        return FindNextVertexNode(vertexID, vf, nameID, typeID, nodeID);
    case E4_VMNODERANDOM:
        return FindNextVertexNodeRandom(vertexID, vf, nameID, typeID, nodeID);
    case E4_VMUNKNOWN:
    case E4_VMLASTMETHOD:
    default:
        return NULL;
    }
}

/*
 * Helper method to visit all vertices of a storage according to some given
 * constraints.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorage(int vertexID, int vf, int nameID,
					     e4_VertexType typeID) const
{
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    switch (vf) {
    case E4_VFNONE:
        return FindNextVertexStorageNone(vertexID);
    case E4_VFNAME:
        return FindNextVertexStorageName(vertexID, nameID);
    case E4_VFTYPE:
        return FindNextVertexStorageType(vertexID, typeID);
    case (E4_VFNAME | E4_VFTYPE):
        return FindNextVertexStorageBoth(vertexID, nameID, typeID);
    default:
        return NULL;
    }
}

/*
 * Helper method to visit all vertices of a storage.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorageNone(int vertexID) const
{
    int i, l;
    
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if (((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) {
	    if ((((int) pFlags(vertices[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedVertex(i)) {
		continue;
	    }
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices of a storage with a given name.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorageName(int vertexID, 
						 int nameID) const
{
    int i, l;
    
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
	    ((int) pNameID(vertices[i]) == nameID)) {
	    if ((((int) pFlags(vertices[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedVertex(i)) {
		continue;
	    }
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices of a storage with a given type.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorageType(int vertexID,
						 e4_VertexType vt) const
{
    int i, l;
    
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
	    if ((((int) pFlags(vertices[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedVertex(i)) {
		continue;
	    }
            return GetVertex(i);
        }
    }

    return NULL;
}

/*
 * Helper method to visit all vertices of a storage with a given name and type.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexStorageBoth(int vertexID,
						 int nameID,
						 e4_VertexType vt) const
{
    int i, l;
    
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((int) pNameID(vertices[i]) == nameID) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
	    if ((((int) pFlags(vertices[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedVertex(i)) {
		continue;
	    }
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices in a node.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNode(int vertexID, int vf, int nameID,
					  e4_VertexType vt, int nodeID) const
{
    switch (vf) {
    case E4_VFNONE:
        return FindNextVertexNodeNone(vertexID, nodeID);
    case E4_VFNAME:
        return FindNextVertexNodeName(vertexID, nameID, nodeID);
    case E4_VFTYPE:
        return FindNextVertexNodeType(vertexID, vt, nodeID);
    case (E4_VFNAME | E4_VFTYPE):
        return FindNextVertexNodeBoth(vertexID, nameID, vt, nodeID);
    default:
        return NULL;
    }
}

/*
 * Helper method to visit all vertices in a node without constraints.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeNone(int vertexID,
                                             int nodeID) const
{
    vertexID = FindNextVertexIndexInNode(vertexID, nodeID);
    if (vertexID == E4_NEXTNONE) {
        return NULL;
    }
    return GetVertex(vertexID);
}

/*
 * Helper method to visit all vertices in a node with a given name.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeName(int vertexID, int nameID,
					      int nodeID) const
{
    int i;
    
    for (i = FindNextVertexIndexInNode(vertexID, nodeID);
         i != E4_NEXTNONE;
         i = (int) pNext(vertices[i])) {
        if ((int) pNameID(vertices[i]) == nameID) {
	    return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices in a node with a given type.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeType(int vertexID, e4_VertexType vt,
					      int nodeID) const
{
    int i;

    for (i = FindNextVertexIndexInNode(vertexID, nodeID);
         i != E4_NEXTNONE;
         i = (int) pNext(vertices[i])) {
        if ((e4_VertexType) (int) pVertexType(vertices[i]) == vt) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices in a node with a given type and name.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeBoth(int vertexID, int nameID,
					      e4_VertexType vt, 
					      int nodeID) const
{
    int i;
    
    for (i = FindNextVertexIndexInNode(vertexID, nodeID);
         i != E4_NEXTNONE;
         i = (int) pNext(vertices[i])) {
        if (((int) pNameID(vertices[i]) == nameID) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices in a node in some random order.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNodeRandom(int vertexID, int vf,
						int nameID, e4_VertexType vt,
						int nodeID) const
{
    switch (vf) {
    case E4_VFNONE:
        return FindNextVertexNRNone(vertexID, nodeID);
    case E4_VFNAME:
        return FindNextVertexNRName(vertexID, nameID, nodeID);
    case E4_VFTYPE:
        return FindNextVertexNRType(vertexID, vt, nodeID);
    case (E4_VFNAME | E4_VFTYPE):
        return FindNextVertexNRBoth(vertexID, nameID, vt, nodeID);
    default:
        return NULL;
    }
}

/*
 * Helper method to visit all vertices in a node in some random order,
 * with no constraints.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNRNone(int vertexID, int nodeID) const
{
    int i, l;
    
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
	    ((int) pNodeID(vertices[i]) == nodeID)) {
	    if ((((int) pFlags(vertices[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedVertex(i)) {
		continue;
	    }
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices with a given name in a supplied node
 * in some random order.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNRName(int vertexID, int nameID, 
					    int nodeID) const
{
    int i, l;
    
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((int) pNodeID(vertices[i]) == nodeID) &&
            ((int) pNameID(vertices[i]) == nameID)) {
	    if ((((int) pFlags(vertices[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedVertex(i)) {
		continue;
	    }
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices with a given type in a supplied node
 * in some random order.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNRType(int vertexID, e4_VertexType vt,
					    int nodeID) const
{
    int i, l;
    
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
        if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((int) pNodeID(vertices[i]) == nodeID) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
	    if ((((int) pFlags(vertices[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedVertex(i)) {
		continue;
	    }
            return GetVertex(i);
        }
    }
    return NULL;
}

/*
 * Helper method to visit all vertices with a given name and type in a supplied
 * node in some random order.
 */

e4_VertexImpl *
e4_MetakitStorageImpl::FindNextVertexNRBoth(int vertexID, int nameID, 
					    e4_VertexType vt, int nodeID) const
{
    int i, l;
    
    if (vertexID == E4_VERTEXNOTFOUND) {
        vertexID = -1;
    }
    for (i = vertexID + 1, l = vertices.GetSize(); i < l; i++) {
	if ((((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) &&
            ((int) pNodeID(vertices[i]) == nodeID) &&
            ((int) pNameID(vertices[i]) == nameID) &&
            ((e4_VertexType) (int) pVertexType(vertices[i]) == vt)) {
	    if ((((int) pFlags(vertices[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedVertex(i)) {
		continue;
	    }
            return GetVertex(i);
        }
    }
    return NULL;
}
    
/*
 * Helper method to find the vertexID with the next rank after a given
 * vertexID in a node identified by a supplied nodeID.
 */

int
e4_MetakitStorageImpl::FindNextVertexIndexInNode(int vertexID, 
						 int nodeID) const
{
    if (vertexID == E4_VERTEXNOTFOUND) {
        if (nodeID == E4_NODENOTFOUND) {
            return E4_NEXTNONE;
        }
        return pFirstVertex(nodes[nodeID]);
    }
    if (((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0) {
        return E4_NEXTNONE;
    }
    if ((int) pNodeID(vertices[vertexID]) != nodeID) {
	return E4_NEXTNONE;
    }
    return (int) pNext(vertices[vertexID]);
}

/*
 * Given a node ID, retrieve the instance of the next node after
 * the given one in this storage.
 */

e4_NodeImpl *
e4_MetakitStorageImpl::FindNextNode(int nodeID) const
{
    int i, l;
    e4_NodeImpl *n;

    if (nodeID == E4_NODENOTFOUND) {
	nodeID = -1;
    }
    for (l = nodes.GetSize(), i = nodeID + 1;
	 i < l;
	 i++) {
	if (((int) pFlags(nodes[i]) & E4_INUSE) == E4_INUSE) {

	    /*
	     * If the node is detached and not referenced by the program,
	     * and it's not the root node, then skip this node.
	     */

	    if ((((int) pFlags(nodes[i]) & E4_DETACHED) == E4_DETACHED) &&
		!IsReferencedNode(i) &&
		(i != (int) pFirst(unused[E4_GRAPHROOTNODE]))) {
		continue;
	    }

	    (void) GetNode(i, n);
	    return n;
	}
    }
    return NULL;
}

/*
 * Given a node ID, retrieve the number of vertices in that node.
 */

int 
e4_MetakitStorageImpl::VertexCountFromNodeID(int nodeID) const
{
    return (int) pVertexCount(nodes[nodeID]);
}

/*
 * Given a node ID, retrieve the number of vertices in that node that
 * have the given name.
 */

int
e4_MetakitStorageImpl::VertexCountWithNameFromNodeID(int nodeID,
						     const char *nm) const
{
    int i, count;
    int nameID = InternName(nm);

    for (i = (int) pFirstVertex(nodes[nodeID]), count = 0;
	 i != E4_NEXTNONE;
	 i = (int) pNext(vertices[i])) {
	if ((int) pNameID(vertices[i]) == nameID) {
	    count++;
	}
    }

    return count;
}

/*
 * Given a node ID, retrieve the number of vertices in that node with the
 * given type.
 */

int
e4_MetakitStorageImpl::VertexCountWithTypeFromNodeID(int nodeID,
						     e4_VertexType tp) const
{
    int i, count;

    for (i = (int) pFirstVertex(nodes[nodeID]), count = 0;
	 i != E4_NEXTNONE;
	 i = (int) pNext(vertices[i])) {
	if ((int) pVertexType(vertices[i]) == tp) {
	    count++;
	}
    }

    return count;
}

/*
 * Given a node ID, retrieve the number of vertices in that node with the
 * same value.
 */

int
e4_MetakitStorageImpl::VertexCountWithValueFromNodeID(int nodeID,
						      e4_Value &v) const
{
    int i, count;

    switch (v.vertexType) {
    case E4_VTNODE:
	{
	    int vn = GetNodeID(v.n);

	    for (i = (int) pFirstVertex(nodes[nodeID]), count = 0;
		 i != E4_NEXTNONE;
		 i = (int) pNext(vertices[i])) {
		if (((int) pVertexType(vertices[i]) == E4_VTNODE) &&
		    ((int) pRowID(vertices[i]) == vn)) {
		    count++;
		}
	    }
	    break;
	}
    case E4_VTINT:
	{
	    int vi = v.u.i;

	    for (i = (int) pFirstVertex(nodes[nodeID]), count = 0;
		 i != E4_NEXTNONE;
		 i = (int) pNext(vertices[i])) {
		if (((int) pVertexType(vertices[i]) == E4_VTINT) &&
		    ((int) pRowID(vertices[i]) == vi)) {
		    count++;
		}
	    }
	    break;
	}
    case E4_VTDOUBLE:
	{
	    double vd = v.u.d;

	    for (i = (int) pFirstVertex(nodes[nodeID]), count = 0;
		 i != E4_NEXTNONE;
		 i = (int) pNext(vertices[i])) {
		if (((int) pVertexType(vertices[i]) == E4_VTDOUBLE) &&
		    (pDoubleVal(doubles[(int) pRowID(vertices[i])]) == vd)) {
		    count++;
		}
	    }
	    break;
	}
    case E4_VTSTRING:
	{
	    const char *vs = v.u.s;

	    for (i = (int) pFirstVertex(nodes[nodeID]), count = 0;
		 i != E4_NEXTNONE;
		 i = (int) pNext(vertices[i])) {
		if (((int) pVertexType(vertices[i]) == E4_VTSTRING) &&
		    (strcmp(pStringVal(strings[(int) pRowID(vertices[i])]), 
			    vs) == 0)) {
		    count++;
		}
	    }
	    break;
	}
    case E4_VTBINARY:
	{
	    const char *vb = (const char *) v.u.b.bytes;
	    const char *sb;
	    int nb = v.u.b.nbytes;
	    int j;

	    for (i = (int) pFirstVertex(nodes[nodeID]), count = 0;
		 i != E4_NEXTNONE;
		 i = (int) pNext(vertices[i])) {
		if ((int) pVertexType(vertices[i]) == E4_VTBINARY) {
		    c4_Bytes bytes = 
		      pBinaryVal(binary[(int)pRowID(vertices[i])]);

		    if (bytes.Size() != nb) {
			continue;
		    }
		    sb = (const char *) (const void *) bytes.Contents();
		    for (j = 0; j < nb; j++) {
			if (sb[j] != vb[j]) {
			    break;
			}
		    }
		    if (j == nb) {
			count++;
		    }
		}
	    }
	    break;
	}
    default:
	count = 0;      
    }

    return count;
}


/*
 * Given a vertex ID retrieve the node containing that vertex.
 */

e4_NodeImpl *
e4_MetakitStorageImpl::ContainingNodeFromVertexID(int vertexID) const 
{
    e4_NodeImpl *n;
    int i;

    i = (int) pNodeID(vertices[vertexID]);
    if (GetNode(pNodeID(vertices[vertexID]), n) == false) {
	return NULL;
    }
    return n;
}

/*
 * Check whether a vertex ID is legal.
 */

bool
e4_MetakitStorageImpl::IsLegalVertexID(int index) const
{
    int flags;

    /*
     * If index is out of bounds, return false.
     */

    if ((index < 0) || (index >= vertices.GetSize())) {
	return false;
    }

    /*
     * If not E4_INUSE, return false.
     */

    flags = (int) pFlags(vertices[index]);
    if ((flags & E4_INUSE) == 0) {
	return false;
    }

    /*
     * If not detached, return true.
     */

    if ((flags & E4_DETACHED) == 0) {
	return true;
    }

    /*
     * If detached, must be referenced by user program to be legal.
     */

    if (!IsReferencedVertex(index)) {
	return false;
    }
    return true;
}

/*
 * Retrieve the vertex type, the ID of the containing node or the name
 * of the vertex, given a vertex ID.
 */

e4_VertexType
e4_MetakitStorageImpl::VertexTypeFromVertexID(int index) const 
{
    return (e4_VertexType) (int) pVertexType(vertices[index]);
}

int 
e4_MetakitStorageImpl::NodeIDFromVertexID(int index) const 
{
    return (int) pNodeID(vertices[index]);
}

const char *
e4_MetakitStorageImpl::VertexNameFromVertexID(int index) const
{
    const char *nm;

    if (GetVertexName(pNameID(vertices[index]), nm) != true) {
	return NULL;
    }
    return nm;
}

/*
 * Find the name ID of a vertex with a given vertex ID.
 */

int
e4_MetakitStorageImpl::NameIDFromVertexID(int vertexID) const 
{
    return (int) pNameID(vertices[vertexID]);
}

/*
 * Rename a vertex given a vertex ID.
 */

bool
e4_MetakitStorageImpl::RenameVertexByVertexID(int vertexID, 
					      const char *newname)
{
    pNameID(vertices[vertexID]) = InternName(newname);
    return true;
}

/*
 * Add the node identified by "parentID" to the list of nodes that
 * contain the node identified by "childID". Add the vertex identified
 * by "vertexID" to the list of vertices referenced by the parent
 * record for "parentID".
 *
 * NOTE: A parent node can contain a child node multiple times, hence we
 * maintain a reference count to efficiently represent this multiple
 * containment.
 */

void
e4_MetakitStorageImpl::AddParent(int childID, int parentID, int vertexID) 
{
    int i, back, pid, flags;

    /*
     * We changed the storage.
     */

    MarkDirty();

    /*
     * The child node is no longer detached.
     */

    flags = (int) pFlags(nodes[childID]);
    flags &= ~(E4_DETACHED | E4_DETACHNOTIFY);
    pFlags(nodes[childID]) = flags;

    /*
     * If the new parentID is E4_NEXTNONE, it means that we're adding
     * a detached vertex to the list of detached vertices whose value
     * is the child node.
     */

    if (parentID == E4_NEXTNONE) {
	pNextInParent(vertices[vertexID]) = 
	    (int) pDetachedVertices(nodes[childID]);
	pDetachedVertices(nodes[childID]) = vertexID;
	return;
    }

    /*
     * Increment the reference count for the child node because
     * it gets a new parent.
     */

    pRefCount(nodes[childID]) = (int) pRefCount(nodes[childID]) + 1;

    /*
     * See if the parentID is already present in the list
     * of parents for the node denoted by childID. In that case
     * increment the number of times the node is a child of this
     * parent, and return.
     */

    for (i = (int) pParentID(nodes[childID]);
	 i != E4_NEXTNONE;
	 i = (int) pNext(parents[i])) {
	if ((int) pNodeID(parents[i]) == parentID) {
	    pCount(parents[i]) = (int) pCount(parents[i]) + 1;

	    /*
	     * Add this vertex to the chain of vertices referenced
	     * by this parent record.
	     */

	    pNextInParent(vertices[vertexID]) = (int) pVertexChain(parents[i]);
	    pVertexChain(parents[i]) = vertexID;

	    return;
	}
    }

    /*
     * Did not find the "parentID" in the current list, so
     * add a new row in the "parents" view for this node.
     */

    if ((int) pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) == E4_NEXTNONE) {
	MakeParentSpace();
    }

    i = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]);
    pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) = (int) pNext(parents[i]);

    /*
     * Must add the new parent as the last one in the list
     * of parents for the child node.
     */

    pFlags(parents[i]) = E4_INUSE;
    pNodeID(parents[i]) = parentID;
    pCount(parents[i]) = 1;
    pNext(parents[i]) = E4_NEXTNONE;
    pVertexChain(parents[i]) = vertexID;

    /*
     * Mark this vertex as the last one in the parent that has as its
     * value this child node.
     */

    pNextInParent(vertices[vertexID]) = E4_NEXTNONE;

    /*
     * After the following loop, back will contain the last parent
     * in the chain of parents for the child node.
     */

    for (back = E4_NEXTNONE, pid = (int) pParentID(nodes[childID]);
	 pid != E4_NEXTNONE;
	 pid = (int) pNext(parents[pid])) {
	back = pid;
    }

    /*
     * If back == E4_NEXTNONE it means that the child node does
     * not yet have any parents, hence make this entry the first
     * one. Otherwise, modify the chain pointer of back to point
     * to the new entry.
     */

    if (back == E4_NEXTNONE) {
	pParentID(nodes[childID]) = i;
    } else {
	pNext(parents[back]) = i;
    }
}

/*
 * Remove a parent from the list of parents of a child, and if requested,
 * decrement the reference count for the child.
 */

void
e4_MetakitStorageImpl::RemoveParent(int childID, int parentID, int vertexID,
				    bool GCOrDetach)
{
    int cur, prev, vv, pv, flags;

    /*
     * Sanity checks:
     */

    if ((childID < 0) ||
	(childID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[childID]) & E4_INUSE) == 0) ||
	(parentID >= nodes.GetSize())) {
	return;
    }

    /*
     * We are making changes to the storage, so mark it dirty.
     */

    MarkDirty();

    /*
     * If the parentID == E4_NEXTNONE then the vertex is a detached one.
     * In that case, splice it out of the list of detached vertices that
     * point at this node.
     */

    if (parentID == E4_NEXTNONE) {

	/*
	 * The child node loses a detached vertex pointing at it.
	 */

	for (prev = E4_NEXTNONE, cur = (int) pDetachedVertices(nodes[childID]);
	     cur != E4_NEXTNONE;
	     cur = (int) pNextInParent(vertices[cur])) {
	    if (cur == vertexID) {
		if (prev == E4_NEXTNONE) {
		    pDetachedVertices(nodes[childID]) =
			(int) pNextInParent(vertices[cur]);
		} else {
		    pNextInParent(vertices[prev]) = 
			(int) pNextInParent(vertices[cur]);
		}
		break;
	    }
	    prev = cur;
	}

	/*
	 * See if we found the vertex among the detached ones that
	 * point at this node.
	 */

	if (cur == E4_NEXTNONE) {
	    return;
	}

    } else {

	/*
	 * Decrement the reference count for the child node
	 * because it loses a parent.
	 */

	pRefCount(nodes[childID]) = (int) pRefCount(nodes[childID]) - 1;
	if ((int) pRefCount(nodes[childID]) < 0) {
	    pRefCount(nodes[childID]) = 0;
	}

	/*
	 * If the parent ID is present in the list of parents of the
	 * child, retrieve the parent record. Otherwise, bail out.
	 */

	for (cur = (int) pParentID(nodes[childID]), prev = E4_NEXTNONE;
	     cur != E4_NEXTNONE;
	     cur = (int) pNext(parents[cur])) {
	    if ((int) pNodeID(parents[cur]) == parentID) {
		break;
	    }
	    prev = cur;
	}
	if (cur == E4_NEXTNONE) {
	    return;
	}

	/*
	 * If the number of times the child node appears in the parent
	 * node is greater than one, decrement the count, and remove the
	 * first vertex in the parent that has the child as its value.
	 */

	if ((int) pCount(parents[cur]) > 1) {
	    pCount(parents[cur]) = (int) pCount(parents[cur]) - 1;

	    /*
	     * Splice the vertex out of the list of vertices in the parent
	     * whose value is this child node.
	     */

	    for (pv = E4_NEXTNONE, vv = (int) pVertexChain(parents[cur]);
		 vv != E4_NEXTNONE;
		 vv = pNextInParent(vertices[vv])) {

		/*
		 * When we find the vertex splice it out of the list
		 * and break.
		 */

		if (vv == vertexID) {
		    if (pv == E4_NEXTNONE) {
			pVertexChain(parents[cur]) =
			  (int) pNextInParent(vertices[vv]);
		    } else {
			pNextInParent(vertices[pv]) =
			  (int) pNextInParent(vertices[vv]);
		    }

		    break;
		}

		pv = vv;
	    }
	    if (vv == E4_NEXTNONE) {
		return;
	    }

	} else {

	    /*
	     * No, this parent had only one reference to this child, so we must
	     * splice out the record from the list of parents. If this is the
	     * first record, then update the pParentID vertex in the child
	     * node to the next record, otherwise splice it out of the list. In
	     * either case, return the record to the freelist of 
	     * parent records.
	     *
	     * NOTE: We dont bother to splice out the vertex from the list of
	     * vertices in the parent record, because the parent record is
	     * recycled anyway. But we must check that indeed the child is
	     * referenced from the parent through the given vertex.
	     */

	    if (prev == E4_NEXTNONE) {
		pParentID(nodes[childID]) = (int) pNext(parents[cur]);
	    } else {
		pNext(parents[prev]) = (int) pNext(parents[cur]);
	    }
	    AddToUnusedParentList(cur);
	}
    }

    /*
     * See if the child node is newly detached (wasn't detached and
     * became detached as a result of this operation).
     */

    if (((int) pParentID(nodes[childID]) == E4_NEXTNONE) &&
	(((int) pFlags(nodes[childID]) & E4_DETACHED) == 0)) {
	flags = (int) pFlags(nodes[childID]);
	flags |= E4_DETACHED;
	flags &= ~(E4_DETACHNOTIFY);
	pFlags(nodes[childID]) = flags;

	/*
	 * See if the node is recycle-able.
	 */

	if (((int) pDetachedVertices(nodes[childID]) == E4_NEXTNONE) &&
	    !IsReferencedNode(childID)) {

	    /*
	     * The node is unreachable. Recycle it.
	     */

	    if (GCOrDetach && !IsGCDeferred()) {
		DoGC();
	    } else {
		needsGC = true;
	    }
	} else if (GCOrDetach) {
	    /*
	     * The node is referenced by the program or by some detached
	     * vertices. Cause a detach event.
	     */

	    flags = (int) pFlags(nodes[childID]);
	    flags |= E4_DETACHNOTIFY;
	    pFlags(nodes[childID]) = flags;

	    if (HasCallbacks(E4_CBDETNODE)) {
		CallbackEvent(E4_CBDETNODE, FindNode(childID));
	    }
	}
    }
}

/*
 * Add entities to their respective free lists.
 */

void
e4_MetakitStorageImpl::AddToUnusedNodesList(int i) 
{
    bool reallyFreed = false;

    if (((int) pFlags(nodes[i]) & E4_INUSE) == E4_INUSE) {
	reallyFreed = true;
    }
    pNext(nodes[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]);
    pFlags(nodes[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) = i;

    if (reallyFreed) {
	statistics[(int) E4_SPNODE][(int) E4_SSUSED]--;
	statistics[(int) E4_SPNODE][(int) E4_SSFREED]++;
    }
}

void
e4_MetakitStorageImpl::AddToUnusedVerticesList(int i) 
{
    bool reallyFreed = false;

    if (((int) pFlags(vertices[i]) & E4_INUSE) == E4_INUSE) {
	reallyFreed = true;
    }
    pNext(vertices[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]);
    pFlags(vertices[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) = i;

    if (reallyFreed) {
	statistics[(int) E4_SPVERTEX][(int) E4_SSUSED]--;
	statistics[(int) E4_SPVERTEX][(int) E4_SSFREED]++;
    }
}

void
e4_MetakitStorageImpl::AddToUnusedIntsList(int i)
{
    /*
     * Nothing to do here.
     */
}

void 
e4_MetakitStorageImpl::AddToUnusedDoublesList(int i) 
{
    pNext(doubles[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]);
    pFlags(doubles[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDDOUBLE]) = i;

    statistics[(int) E4_SPDOUBLE][(int) E4_SSUSED]--;
    statistics[(int) E4_SPDOUBLE][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedStringsList(int i) 
{
    pNext(strings[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]);
    pFlags(strings[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDSTRING]) = i;

    statistics[(int) E4_SPSTRING][(int) E4_SSUSED]--;
    statistics[(int) E4_SPSTRING][(int) E4_SSFREED]++;
}

void 
e4_MetakitStorageImpl::AddToUnusedNamesList(int i) 
{
    pNext(names[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]);
    pFlags(names[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDNAME]) = i;

    statistics[(int) E4_SPNAME][(int) E4_SSUSED]--;
    statistics[(int) E4_SPNAME][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedBinaryList(int i)
{
    pNext(binary[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]);
    pFlags(binary[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDBINARY]) = i;

    statistics[(int) E4_SPBINARY][(int) E4_SSUSED]--;
    statistics[(int) E4_SPBINARY][(int) E4_SSFREED]++;
}

void
e4_MetakitStorageImpl::AddToUnusedParentList(int i)
{
    pNext(parents[i]) = (int) pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]);
    pFlags(parents[i]) = 0;
    pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) = i;
}

/*
 * Move the vertex indentified by vertexID into the node identified by
 * nodeID at the given rank.
 */

bool
e4_MetakitStorageImpl::MoveVertex(int vertexID, int nodeID, int rank)
{
    if ((vertexID < 0) ||
	(vertexID >= vertices.GetSize()) ||
	(nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0) || 
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0) ||
	(rank == 0)) {
	return false;
    }
    if (rank > 1) {
	/*
	 * If we are trying to insert beyond the current end of the
	 * node, prevent it.
	 */

	if ((int) pVertexCount(nodes[nodeID]) < rank - 1) {
	    return false;
	}

	/*
	 * If the requested rank is exactly the new last vertex in
	 * the node, convert it to the "insert as last" case.
	 */

	if ((int) pVertexCount(nodes[nodeID]) == rank - 1) {
	    rank = -1;
	}
    }

    /*
     * Move the vertex.
     */

    SpliceIn(vertexID, SpliceOut(vertexID, false), nodeID, rank);

    return true;
}

/*
 * Helper functions for moving vertices between nodes.
 */

int
e4_MetakitStorageImpl::SpliceOut(int vertexID, bool updateChild)
{
    int nodeID = (int) pNodeID(vertices[vertexID]);
    int nextID, prevID;

    /*
     * Splice the vertex out of its containing node.
     */

    nextID = (int) pNext(vertices[vertexID]);
    prevID = (int) pPrev(vertices[vertexID]);

    if (nodeID != E4_NEXTNONE) {
	if ((int) pFirstVertex(nodes[nodeID]) == vertexID) {
	    pFirstVertex(nodes[nodeID]) = nextID;
	}
	if ((int) pLastVertex(nodes[nodeID]) == vertexID) {
	    pLastVertex(nodes[nodeID]) = prevID;
	}

	/*
	 * Decrement the vertex count for the node out of which we spliced
	 * the vertex.
	 */

	pVertexCount(nodes[nodeID]) = (int) pVertexCount(nodes[nodeID]) - 1;
    }

    if (nextID != E4_NEXTNONE) {
	pPrev(vertices[nextID]) = prevID;
    }
    if (prevID != E4_NEXTNONE) {
	pNext(vertices[prevID]) = nextID;
    }

    /*
     * If the value of this vertex is a node and updateChild == true, 
     * we must splice the vertex out of the list of vertices in the
     * parent for which this node is the value.
     */

    if ((updateChild) && 
	((e4_VertexType) (int) pVertexType(vertices[vertexID]) == E4_VTNODE)) {
	RemoveParent((int) pRowID(vertices[vertexID]), nodeID, vertexID, true);
    }

    /*
     * This vertex now is not part of any node.
     */

    pNodeID(vertices[vertexID]) = E4_NEXTNONE;
    pNextInParent(vertices[vertexID]) = E4_NEXTNONE;
    pPrev(vertices[vertexID]) = E4_NEXTNONE;
    pNext(vertices[vertexID]) = E4_NEXTNONE;

    /*
     * Finally return the nodeID for the parent from which we just
     * spliced this vertex out.
     */

    return nodeID;
}

void
e4_MetakitStorageImpl::SpliceIn(int vertexID, int oldParent,
				int newParent, int rank)
{
    int cnt, childID, nextID, prevID, flags;

    switch (rank) {
    case 1:
	/*
	 * Insert the new vertex as the first vertex in this node.
	 */

	pNext(vertices[vertexID]) = nextID = 
	  (int) pFirstVertex(nodes[newParent]);
	if (nextID != E4_NEXTNONE) {
	    pPrev(vertices[nextID]) = vertexID;
	}
	pPrev(vertices[vertexID]) = E4_NEXTNONE;
	if ((int) pLastVertex(nodes[newParent]) == E4_NEXTNONE) {
	    pLastVertex(nodes[newParent]) = vertexID;
	}
	pFirstVertex(nodes[newParent]) = vertexID;
	break;

    case -1: 
	/* 
	 * Insert the new vertex as the last vertex in this node.
	 */

	pNext(vertices[vertexID]) = E4_NEXTNONE;
	pPrev(vertices[vertexID]) = prevID = 
	  (int) pLastVertex(nodes[newParent]);
	if (prevID == E4_NEXTNONE) {
	    pFirstVertex(nodes[newParent]) = vertexID;
	} else {
	    pNext(vertices[prevID]) = vertexID;
	}
	pLastVertex(nodes[newParent]) = vertexID;
	break;

    default:
	/*
	 * Insert the new vertex at the specified rank position.
	 */

	for (prevID = (int) pFirstVertex(nodes[newParent]), cnt = 1;
	     cnt < rank - 1;
	     prevID = (int) pNext(vertices[prevID]), cnt++) {
	}
	pNext(vertices[vertexID]) = nextID = (int) pNext(vertices[prevID]);
	pPrev(vertices[vertexID]) = prevID;
	if (nextID != E4_NEXTNONE) {
	    pPrev(vertices[nextID]) = vertexID;
	}
	pNext(vertices[prevID]) = vertexID;
	break;
    }

    /*
     * Special handling if the vertex contains a node: in that case
     * we remove the old parent and add the new parent, without
     * letting the user program notice if the node became temporarily
     * detached.
     */

    if ((oldParent != newParent) &&
	((e4_VertexType) (int) pVertexType(vertices[vertexID]) == E4_VTNODE)) {
	childID = (int) pRowID(vertices[vertexID]);
	RemoveParent(childID, oldParent, vertexID, false);
	AddParent(childID, newParent, vertexID);
    }

    /*
     * Record that the vertex is now in newParent.
     */

    pNodeID(vertices[vertexID]) = newParent;

    /*
     * The moved vertex is no longer detached.
     */

    flags = (int) pFlags(vertices[vertexID]);
    flags = ~(E4_DETACHED | E4_DETACHNOTIFY);
    pFlags(vertices[vertexID]) = flags;

    /*
     * Finally increment the vertex count in this node to account for
     * the newly added vertex.
     */

    pVertexCount(nodes[newParent]) = (int) pVertexCount(nodes[newParent]) + 1;
}

/*
 * Copy all e4Graph data from this storage to another storage.
 */

bool
e4_MetakitStorageImpl::CopyTo(e4_StorageImpl *osp)
{
    e4_MetakitStorageImpl *omsp;
    int i, l;

    /*
     * Check the other storage is using the Metakit storage driver. If
     * not, bail out.
     */

    if (strcmp(GetDriver(), E4_METAKIT) != 0) {
	return false;
    }
    omsp = (e4_MetakitStorageImpl *) osp;

    /*
     * Empty out the cache of the other storage.
     */

    omsp->CleanUp();

    /*
     * Copy all tables from this storage to the other storage.
     */

    omsp->nodes.SetSize(nodes.GetSize());
    omsp->vertices.SetSize(vertices.GetSize());
    omsp->doubles.SetSize(doubles.GetSize());
    omsp->strings.SetSize(strings.GetSize());
    omsp->binary.SetSize(binary.GetSize());
    omsp->names.SetSize(names.GetSize());
    omsp->unused.SetSize(unused.GetSize());
    omsp->parents.SetSize(parents.GetSize());

    for (i = 0, l = nodes.GetSize(); i < l; i++) {
	omsp->nodes[i] = nodes[i];
    }
    for (i = 0, l = vertices.GetSize(); i < l; i++) {
	omsp->vertices[i] = vertices[i];
    }
    for (i = 0, l = doubles.GetSize(); i < l; i++) {
	omsp->doubles[i] = doubles[i];
    }
    for (i = 0, l = strings.GetSize(); i < l; i++) {
	omsp->strings[i] = strings[i];
    }
    for (i = 0, l = binary.GetSize(); i < l; i++) {
	omsp->binary[i] = binary[i];
    }
    for (i = 0, l = names.GetSize(); i < l; i++) {
	omsp->names[i] = names[i];
    }
    for (i = 0, l = unused.GetSize(); i < l; i++) {
	omsp->unused[i] = unused[i];
    }
    for (i = 0, l = parents.GetSize(); i < l; i++) {
	omsp->parents[i] = parents[i];
    }

    /*
     * Finally initialize the new data in the other storage.
     */

    omsp->Initialize();

    return true;
}

/*
 * Destroy the physical storage and make this instance invalid.
 */

void
e4_MetakitStorageImpl::Destroy()
{
    /*
     * First of all, clean up the state associated with this storage.
     */

    CleanUp();

    /*
     * Destroying a Metakit storage is a two step process. First
     * we have to discard the last reference to the storage, to
     * give Metakit a chance to commit the storage. Then, we unlink
     * the file containing the storage so that it will no longer
     * be present on the disk. The order of these operations is
     * important: if you reverse the order the effect of unlinking
     * the file will be undone when Metakit commits the file and
     * makes it appear again!
     */

    if (storage != NULL) {
	delete storage;
	storage = NULL;
    }
    if (name != NULL) {
	unlink(name);
    }
}

/*
 * How many parents does the node with the given ID have?
 */

int
e4_MetakitStorageImpl::ParentCount(int nodeID) const
{
    int count, pid;

    if ((nodeID < 0) || (nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return E4_NODENOTFOUND;
    }

    /*
     * Count the number of different nodes that are parents of
     * the node with the given node ID.
     */

    for (count = 0, pid = (int) pParentID(nodes[nodeID]);
	 pid != E4_NEXTNONE;
	 count++, pid = (int) pNext(parents[pid]));

    return count;
}

/*
 * How many vertices contain the node with the given nodeID as their value?
 */

int
e4_MetakitStorageImpl::OccurrenceCount(int nodeID) const
{
    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return E4_NODENOTFOUND;
    }
    return (int) pRefCount(nodes[nodeID]);
}

/*
 * How many vertices in the node identified by the given parent ID contain
 * as their value the node denoted by the child ID?
 */

int
e4_MetakitStorageImpl::OccurrenceCount(int nodeID, int parentID) const
{
    int pid;

    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0) ||
	(parentID < 0) || 
	(parentID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[parentID]) & E4_INUSE) == 0)) {
	return E4_NODENOTFOUND;
    }

    /*
     * Search for the parent record for the child node that contains
     * the parent node information:
     */

    for (pid = (int) pParentID(nodes[nodeID]);
	 pid != E4_NEXTNONE;
	 pid = (int) pNext(parents[pid])) {
	if ((int) pNodeID(parents[pid]) == parentID) {
	    return (int) pCount(parents[pid]);
	}
    }

    /*
     * Didn't find the parent in the list of parents for this child. So
     * the child occurs zero times in that parent.
     */

    return 0;
}

/*
 * What is the rank of the parent node with the given parent ID for the
 * child node with the given child ID?
 */

int
e4_MetakitStorageImpl::ParentRank(int nodeID, int parentID) const
{
    int pid, rank;

    if ((nodeID < 0) ||
	(nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0) ||
	(parentID < 0) || 
	(parentID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[parentID]) & E4_INUSE) == 0)) {
	return E4_NODENOTFOUND;
    }

    /*
     * Search for the parent record for the child node that contains
     * the parent node information:
     */

    for (pid = (int) pParentID(nodes[nodeID]), rank = 1;
	 pid != E4_NEXTNONE;
	 pid = (int) pNext(parents[pid]), rank++) {
	if ((int) pNodeID(parents[pid]) == parentID) {
	    return rank;
	}
    }

    return E4_NODENOTFOUND;
}

/*
 * Return the hash code for this storage.
 */

int
e4_MetakitStorageImpl::HashCode() const
{
    return (int) pFirst(unused[E4_GRAPHHASHCODE]);
}

/*
 * Get a value for a statistic being collected for this storage.
 */

bool
e4_MetakitStorageImpl::GetStatistic(e4_Space sp, e4_SpaceStat st, int &v)
{
    /*
     * Select the statistic to retrieve.
     */

    if (((int) sp < 0) || ((int) sp >= (int) E4_SPLAST) ||
	((int) st < 0) || ((int) st >= (int) E4_SSLAST)) {
	return false;
    }
    v = statistics[(int) sp][(int) st];
    return true;
}

/*
 * Compute the number of slots available in a specific space in the storage.
 */

void
e4_MetakitStorageImpl::UpdateAvailable(e4_Space sp)
{
    /* TODO: Implement! */
}

/*
 * Compute the number of slots used in a specific space in the storage.
 */

void
e4_MetakitStorageImpl::UpdateUsed(e4_Space sp)
{
    /* TODO: Implement! */
}

/*
 * DETACHMENT MECHANISM:
 * =====================
 */

/*
 * Detach the node identified by the given nodeID from all 
 * vertices that have this node as their value.
 */

bool
e4_MetakitStorageImpl::DetachNodeByID(int nodeID)
{
    int nstate, pid, nextParentID, vertexID, nextVertexID;

    /*
     * Sanity checks.
     */

    if ((nodeID < 0) || (nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return false;
    }

    /*
     * If already detached, punt.
     */

    if (((int) pFlags(nodes[nodeID]) & E4_DETACHED) == E4_DETACHED) {
	return true;
    }

    /*
     * Iterate over all parents and detach all vertices that have the
     * child node as their value.
     */

    for (pid = (int) pParentID(nodes[nodeID]);
	 pid != E4_NEXTNONE;
	 pid = nextParentID) {

	/*
	 * Iterate over the vertices and set each of them to {E4_VTINT, 0}.
	 */

	for (vertexID = (int) pVertexChain(parents[pid]);
	     vertexID != E4_NEXTNONE;
	     vertexID = nextVertexID) {

	    /*
	     * This vertex no longer denotes the nodeID.
	     */

	    nextVertexID = (int) pNextInParent(vertices[vertexID]);
	    pNextInParent(vertices[vertexID]) = E4_NEXTNONE;

	    /*
	     * Detach this vertex from its parent.
	     */

	    DetachVertexInternal(vertexID, nodeID);
	}

        /*
	 * Recycle this parent record.
	 */
  
	nextParentID = (int) pNext(parents[pid]);
	AddToUnusedParentList(pid);
    }

    /*
     * This node now has no parents.
     */

    pRefCount(nodes[nodeID]) = 0;
    pParentID(nodes[nodeID]) = E4_NEXTNONE;

    /*
     * Mark the node as detached.
     */

    nstate = (int) pFlags(nodes[nodeID]);
    nstate |= (E4_DETACHED | E4_DETACHNOTIFY);
    pFlags(nodes[nodeID]) = nstate;

    /*
     * If the node is referenced, mark it as detached and cause callbacks
     * to be called. Otherwise, delete it from the storage.
     */

    if (IsReferencedNode(nodeID)) {

	/*
	 * Cause callbacks to be called for this detach event.
	 */

	if (HasCallbacks(E4_CBDETNODE)) {
	    CallbackEvent(E4_CBDETNODE, FindNode(nodeID));
	}
    } else {

	/*
	 * Note that there's recycle-able memory.
	 */

	needsGC = true;
    }

    /*
     * If a GC is needed and it's not deferred, do it now.
     */

    if (needsGC && !IsGCDeferred()) {
	DoGC();
    }

    /*
     * We modified the storage.
     */

    MarkDirty();

    return true;
}

/*
 * Quick, internal version of vertex detachment. This version is
 * called by DetachNodeByID. It just detaches the vertex from
 * inside the node that contains it (if any) and updates that
 * node's vertex count.
 *
 * This version also does no sanity checks.
 */

void
e4_MetakitStorageImpl::DetachVertexInternal(int vertexID, int childID)
{
    int prev, next, nodeID, flags;

    nodeID = (int) pNodeID(vertices[vertexID]);

    /*
     * If for whatever reason the vertex is not inside any node,
     * just punt.
     */

    if (nodeID == E4_NEXTNONE) {
	return;
    }

    /*
     * Splice this vertex out of the list of vertices in
     * the containing node:
     */

    next = (int) pNext(vertices[vertexID]);
    prev = (int) pPrev(vertices[vertexID]);

    if (prev == E4_NEXTNONE) {
	pFirstVertex(nodes[nodeID]) = next;
    } else {
	pNext(vertices[prev]) = next;
    }
    if (next == E4_NEXTNONE) {
	pLastVertex(nodes[nodeID]) = prev;
    } else {
	pPrev(vertices[next]) = prev;
    }

    /*
     * Update the vertex count for the node that used to
     * contain this vertex.
     */

    pVertexCount(nodes[nodeID]) = (int) pVertexCount(nodes[nodeID]) - 1;

    /*
     * Really detach this vertex:
     */

    flags = (int) pFlags(vertices[vertexID]);
    flags |= (E4_DETACHED | E4_DETACHNOTIFY);
    pFlags(vertices[vertexID]) = flags;

    pNodeID(vertices[vertexID]) =
      pNextInParent(vertices[vertexID]) =
      pPrev(vertices[vertexID]) =
      pNext(vertices[vertexID]) = E4_NEXTNONE;

    /*
     * If the vertex is referenced from the user program, move this
     * vertex into the list of detached vertices that point at this
     * node. If there are vertex detach callbacks registered, cause
     * a vertex detach event.
     *
     * If the vertex is not referenced, then we just created some
     * recycle-able storage.
     */

    if (IsReferencedVertex(vertexID)) {

	/*
	 * Move this vertex into the list of detached vertices pointing
	 * at the child node.
	 */

	pNextInParent(vertices[vertexID]) =
	 (int) pDetachedVertices(nodes[childID]);
	pDetachedVertices(nodes[childID]) = (int) vertexID;

	/*
	 * Cause a detach event.
	 */

	if (HasCallbacks(E4_CBDETVERTEX)) {
	    CallbackEvent(E4_CBDETVERTEX, GetVertex(vertexID));
	}
    } else {
	needsGC = true;
    }
}

/*
 * Detach this vertex from its containing node.
 */

bool
e4_MetakitStorageImpl::DetachVertexByID(int vertexID)
{
    int vstate;

    /*
     * Sanity checks.
     */

    if ((vertexID < 0) || (vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return false;
    }

    /*
     * If the vertex is already detached, punt.
     */

    if (((int) pFlags(vertices[vertexID]) & E4_DETACHED) == E4_DETACHED) {
	return true;
    }

    /*
     * Splice the vertex out of its old parent. Note that this also
     * takes care of the case where the vertex point at a node, updating
     * the parent bookkeeping for the child node and moving the vertex
     * to the list of detached vertices pointing at the node.
     */

    if ((int) pNodeID(vertices[vertexID]) != E4_NEXTNONE) {
	(void) SpliceOut(vertexID, true);
    }

    /*
     * Mark the vertex as detached.
     */

    vstate = (int) pFlags(vertices[vertexID]);
    vstate |= (E4_DETACHED | E4_DETACHNOTIFY);
    pFlags(vertices[vertexID]) = vstate;

    /*
     * If the vertex is referenced, mark it as detached. Otherwise
     * delete it from the storage.
     */

    if (IsReferencedVertex(vertexID)) {

	/*
	 * Cause callbacks to be called for this detach event.
	 */

	if (HasCallbacks(E4_CBDETVERTEX)) {
	    CallbackEvent(E4_CBDETVERTEX, GetVertex(vertexID));
	}
    } else {
	needsGC = true;
    }

    /*
     * If a GC is needed and GC is not being deferred, then clean up
     * right away.
     */

    if (needsGC && !IsGCDeferred()) {
	DoGC();
    }

    return true;
}

/*
 * Detach the first vertex in parentID that has the node value childID.
 */

bool
e4_MetakitStorageImpl::DetachFirstVertexWithNode(int parentID, int childID)
{
    int i;

    if (storage == NULL) {
	return false;
    }

    /*
     * Find the vertex.
     */

    for (i = (int) pFirstVertex(nodes[parentID]);
	 i != E4_NEXTNONE;
	 i = (int) pNext(vertices[i])) {
	if (((e4_VertexType) (int) pVertexType(vertices[i]) == E4_VTNODE) &&
	    ((int) pRowID(vertices[i]) == childID)) {
	    break;
	}
    }

    /*
     * If we didn't find the vertex, bail out.
     */

    if (i == E4_NEXTNONE) {
	return false;
    }

    /*
     * Found it, so detach this vertex.
     */

    return DetachVertexByID(i);
}

/*
 * Is the node identified by the given ID detached?
 */

bool
e4_MetakitStorageImpl::IsDetachedNodeID(int nodeID)
{
    if ((nodeID < 0) || (nodeID >= nodes.GetSize()) ||
	(((int) pFlags(nodes[nodeID]) & E4_INUSE) == 0)) {
	return false;
    }
    if (((int) pFlags(nodes[nodeID]) & E4_DETACHED) == E4_DETACHED) {
	return true;
    }
    return false;
}

/*
 * Is the vertex identified by the given ID detached?
 */

bool
e4_MetakitStorageImpl::IsDetachedVertexID(int vertexID)
{
    if ((vertexID < 0) || (vertexID >= vertices.GetSize()) ||
	(((int) pFlags(vertices[vertexID]) & E4_INUSE) == 0)) {
	return false;
    }
    if (((int) pFlags(vertices[vertexID]) & E4_DETACHED) == E4_DETACHED) {
	return true;
    }
    return false;
}

/*
 * GC MECHANISM:
 * =============
 */

/*
 * Do a GC.
 */

void
e4_MetakitStorageImpl::DoGC()
{
    bool gcDeferred;

    if (storage == NULL) {
	return;
    }

    /*
     * If we're already deferring GC (perhaps because we're in
     * the middle of a GC) then do not enter GC even if called
     * explicitly. Also do not reset needsGC.
     */

    if (IsGCDeferred()) {
	return;
    }

    /*
     * Defer GC until we're done, and keep looping until needsGC
     * is false.
     */

    gcDeferred = SetGCDeferral(true);
    do {
	InitGC();
	SpanReachableNodes();
	CollectUnreachableEntities();
	FireEventsForNewlyDetached();
    } while (needsGC);
    SetGCDeferral(gcDeferred);
}

/*
 * Helper method to reinitialize the GC counters.
 */

void
e4_MetakitStorageImpl::InitGC()
{
    needsGC = false;
}

/*
 * Is GC deferred?
 */

bool
e4_MetakitStorageImpl::IsGCDeferred() const
{
    if (storage == NULL) {
	return false;
    }
    if ((int) pFirst(unused[E4_GRAPHDEFERGC]) == 1) {
	return true;
    }
    return false;
}

/*
 * Set the GC deferral flag to a new value.
 */

bool
e4_MetakitStorageImpl::SetGCDeferral(bool newval)
{
    int nv = (newval == true) ? 1 : 0;
    bool oldval;

    if (storage == NULL) {
	return false;
    }
    oldval = ((int) pFirst(unused[E4_GRAPHDEFERGC]) == 0) ? false : true;
    pFirst(unused[E4_GRAPHDEFERGC]) = nv;
    return oldval;
}

/*
 * Is GC needed?
 */

bool
e4_MetakitStorageImpl::NeedsGC()
{
    return needsGC;
}

/*
 * Set the flag indicating whether GC is needed.
 */

void
e4_MetakitStorageImpl::SetNeedsGC(bool needs)
{
    needsGC = needs;
}

/*
 * Span all reachable nodes. At the end of this procedure all reachable
 * entities have E4_REACHABLE in their flags.
 */

void
e4_MetakitStorageImpl::SpanReachableNodes()
{
    /*
     * Seed the set of reachable nodes from the exported nodes. This
     * may add some nodes that are exported but detached.
     */

    SeedReachableNodesFromReferencedNodes();

    /*
     * Seed the set of reachable nodes from the exported vertices. This
     * may add some nodes that are reachable only from referenced vertices.
     */

    SeedReachableNodesFromReferencedVertices();

    /*
     * Now span the initial set to mark the complete set of all
     * reachable nodes. As a side effect also mark all reachable vertices.
     */

    SpanSeededNodes();
}

/*
 * Seed the set of reachable nodes from the set of referenced nodes and
 * from the distinguished root node.
 */

void
e4_MetakitStorageImpl::SeedReachableNodesFromReferencedNodes()
{
    int i, l, nstate;

    for (i = 0, l = nodes.GetSize(); i < l; i++) {
	nstate = (int) pFlags(nodes[i]);

	/*
	 * Look for IN_USE nodes.
	 */

	if ((nstate & E4_INUSE) == E4_INUSE) {

	    /*
	     * If this node is referenced by the user program it is
	     * considered to be reachable.
	     */

	    if (IsReferencedNode(i)) {
		idStack1->Push(i);
	    }
	}
    }

    /*
     * Last, seed from the root node.
     */

    idStack1->Push((int) pFirst(unused[E4_GRAPHROOTNODE]));
}

/*
 * Seed the set of reachable nodes from the set of referenced vertices.
 * Some of these vertices may hold the only reference to a node.
 */

void
e4_MetakitStorageImpl::SeedReachableNodesFromReferencedVertices()
{
    int i, l, vstate, nodeID;

    for (i = 0, l = vertices.GetSize(); i < l; i++) {
	vstate = (int) pFlags(vertices[i]);

	/*
	 * Look for IN_USE vertices.
	 */

	if ((vstate & E4_INUSE) == E4_INUSE) {

	    /*
	     * If the vertex is referenced from the user program, it is
	     * considered reachable. Check if the value is a node. If so,
	     * that node is also considered reachable. Do *NOT* mark the
	     * node as reachable here so that the spanning phase knows to
	     * span through it to all reachable nodes from this node.
	     */

	    if (IsReferencedVertex(i)) {
		vstate |= E4_REACHABLE;
		pFlags(vertices[i]) = vstate;

		/*
		 * If the value is a node, that node is also considered
		 * to be reachable.
		 */

		if ((e4_VertexType) (int) pVertexType(vertices[i]) 
		    == E4_VTNODE) {
		    nodeID = (int) pRowID(vertices[i]);
		    idStack1->Push(nodeID);
		}
	    }
	}
    }
}

/*
 * Span the set of nodes collected in the seeding phase, marking each one
 * as reachable, and also marking any vertices that are contained in
 * reachable nodes as reachable.
 */

void
e4_MetakitStorageImpl::SpanSeededNodes()
{
    int nodeID, childID, vertexID, nstate, cstate, vstate;
    bool changed;
    e4_IntStack *tmpstk;

    /*
     * Iterate until changed == false. If a node is found that is not
     * marked as reachable, then:
     * - this node is marked as reachable,
     * - changed is set to true,
     * - the child nodes of this node are added to the spanning set,
     * - the vertices in this node are marked as reachable and
     */

    do {
	/*
	 * No change so far.
	 */

	changed = false;

	/*
	 * Iterate over all nodes currently known as seeds:
	 */

	while (idStack1->Pop(&nodeID)) {

	    /*
	     * Check if this node is not yet known to be reachable.
	     */

	    nstate = (int) pFlags(nodes[nodeID]);
	    if ((nstate & E4_REACHABLE) == 0) {

		/*
		 * Found a node that is not yet known to be reachable.
		 *
		 * Step 1: Mark this node as reachable.
		 */

		nstate |= E4_REACHABLE;
		pFlags(nodes[nodeID]) = nstate;

		/*
		 * Step 2: Record that something changed.
		 */

		changed = true;

		/*
		 * Step 3 and 4: Add the child nodes to the set of seeds
		 * to be examined next, and mark all vertices in this node
		 * as reachable.
		 */

		for (vertexID = (int) pFirstVertex(nodes[nodeID]);
		     vertexID != E4_NEXTNONE;
		     vertexID = (int) pNext(vertices[vertexID])) {

		    vstate = (int) pFlags(vertices[vertexID]);

		    /*
		     * Sanity check: Vertex cannot be detached.
		     */

		    if ((vstate & E4_DETACHED) == E4_DETACHED) {
			fprintf(stderr,
			 "Inconsistent state: detached vertex %d in node %d\n",
			 vertexID, nodeID);
		    }

		    /*
		     * Mark this vertex as reachable.
		     */

		    vstate |= E4_REACHABLE;
		    pFlags(vertices[vertexID]) = vstate;

		    /*
		     * See if the value is a node.
		     */

		    if ((e4_VertexType) (int) pVertexType(vertices[vertexID])
			== E4_VTNODE) {

			/*
			 * Found a node.
			 */

			childID = (int) pRowID(vertices[vertexID]);

			/*
			 * Sanity checks: cannot be detached.
			 */

			cstate = (int) pFlags(nodes[childID]);
			if ((cstate & E4_DETACHED) == E4_DETACHED) {
			  fprintf(stderr,
			 "Inconsistent state: detached node %d in vertex %d\n",
				  childID, vertexID);
			}

			/*
			 * Add the node to the set of seeds to be examined
			 * next round.
			 */

			idStack2->Push(childID);
		    }
		}
	    }
	}

	/*
	 * Switch stacks. At this point idStack2 contains the newly
	 * reachable nodes whose children may not yet have been
	 * explored, and idStack1 is empty.
	 */

	tmpstk = idStack1;
	idStack1 = idStack2;
	idStack2 = tmpstk;

    } while (changed);
}

/*
 * Collect all unreachable entities.
 */

void
e4_MetakitStorageImpl::CollectUnreachableEntities()
{
    /*
     * Order is important: we must collect the unreachable vertices
     * before the unreachable nodes, because vertices whose value is
     * a node are detached from that node. If the node is also not
     * reachable, recycling it first could cause "dangling pointers".
     */

    CollectUnreachableVertices();
    CollectUnreachableNodes();
}

/*
 * Collect all unreachable nodes, add them to the nodes free list.
 */

void
e4_MetakitStorageImpl::CollectUnreachableNodes()
{
    int i, nstate, next, pstate;

    /*
     * Reset nodes free list.
     */

    pFirst(unused[E4_GRAPHFIRSTUNUSEDNODE]) = E4_NEXTNONE;

    /*
     * Add all nodes that aren't explicitly marked as reachable to
     * the nodes free list. Traverse in reverse order to that the
     * free list will be ordered in ascending order. All parent
     * records that are no longer used are marked as such.
     */

    for (i = nodes.GetSize() - 1; i > -1; i--) {
	nstate = (int) pFlags(nodes[i]);

	/*
	 * If the node is not marked as reachable, add it to the
	 * nodes free list. Also mark any attached parent records
	 * as no longer in use if the node was in use.
	 */

	if ((nstate & E4_REACHABLE) == 0) {
	    if ((nstate & E4_INUSE) == E4_INUSE) {
		for (next = (int) pParentID(nodes[i]);
		     next != E4_NEXTNONE;
		     next = (int) pNext(parents[next])) {
		    pFlags(parents[next]) = 0;
		}
		for (next = (int) pFirstVertex(nodes[i]);
		     next != E4_NEXTNONE;
		     next = (int) pNext(vertices[next])) {
		    pNodeID(vertices[next]) = E4_NEXTNONE;
		}
	    }
	    AddToUnusedNodesList(i);
	} else {
	    nstate &= ~(E4_REACHABLE);
	    pFlags(nodes[i]) = nstate;
	}
    }

    /*
     * Now rebuild the parents free list.
     */

    pFirst(unused[E4_GRAPHFIRSTUNUSEDPARENT]) = E4_NEXTNONE;

    for (i = parents.GetSize() - 1; i > -1; i--) {
	pstate = (int) pFlags(parents[i]);
	if ((pstate & E4_INUSE) == 0) {
	    AddToUnusedParentList(i);
	}
    }
}

/*
 * Collect unreachable vertices, add them to the vertices free list and
 * recycle the storage for the value.
 */

void
e4_MetakitStorageImpl::CollectUnreachableVertices()
{
    int i, vstate, nodeID;

    /*
     * Reset vertices free list.
     */

    pFirst(unused[E4_GRAPHFIRSTUNUSEDVERTEX]) = E4_NEXTNONE;

    /*
     * Add all vertices that aren't explicitly marked reachable to the
     * free list. Also recycle the value. Traverse in descending order
     * so that the free list will be ordered in ascending order.
     */

    for (i = vertices.GetSize() - 1; i > -1; i--) {
	vstate = (int) pFlags(vertices[i]);
	if ((vstate & E4_REACHABLE) == 0) {

	    /*
	     * If the vertex was in use, recycle its value too.
	     */

	    if ((vstate & E4_INUSE) == E4_INUSE) {

		/*
		 * Check if the value is a node. If not, just recycle the
		 * value. If it is a node, check if the node is marked
		 * reachable. If it is, splice the vertex out of its
		 * parents, otherwise don't bother.
		 *
		 * If we remove the old parent, do not recursively start
		 * a new GC or callback event processing. Callbacks will
		 * be processed en-masse when the GC is done.		 
		 */

		if ((e4_VertexType) (int) pVertexType(vertices[i]) ==
		    E4_VTNODE) {
		    nodeID = (int) pRowID(vertices[i]);
		    if (nodeID != E4_NEXTNONE) {
			if (((int) pFlags(nodes[nodeID]) & E4_REACHABLE)
			    == E4_REACHABLE) {
			    RemoveParent(nodeID,
					 (int) pNodeID(vertices[i]),
					 i,
					 false);
			}
		    }
		} else {
		    FreeVertexValue(i); 
		}
	    }
	    AddToUnusedVerticesList(i);
	} else {
	    vstate &= ~(E4_REACHABLE);
	    pFlags(vertices[i]) = vstate;
	}
    }
}

/*
 * Fire detach events for entities that became newly detached
 * during this GC.
 */

void
e4_MetakitStorageImpl::FireEventsForNewlyDetached()
{
    FireEventsForNewlyDetachedNodes();
    FireEventsForNewlyDetachedVertices();
}

void
e4_MetakitStorageImpl::FireEventsForNewlyDetachedNodes()
{
    int i, l, flags;
    bool nodeDetCallbacks;

    nodeDetCallbacks = HasCallbacks(E4_CBDETNODE);

    /*
     * We must iterate over all nodes even when there are no
     * registered callbacks, to correctly set the E4_DETACHED
     * flag for newly detached nodes.
     */

    for (i = 0, l = nodes.GetSize(); i < l; i++) {
	flags = (int) pFlags(nodes[i]);
	if ((flags & E4_INUSE) == 0) {
	    continue;
	}

	/*
	 * If this node already has been notified of
	 * a detach, don't bother.
	 */

	if ((flags & E4_DETACHNOTIFY) == E4_DETACHNOTIFY) {
	    continue;
	}

	/*
	 * If we found a detached node, process it.
	 */

	if (((int) pParentID(nodes[i]) == E4_NEXTNONE) &&
	    ((int) pDetachedVertices(nodes[i]) == E4_NEXTNONE)) {
	    flags |= (E4_DETACHED | E4_DETACHNOTIFY);
	    pFlags(nodes[i]) = flags;
	    if (!IsReferencedNode(i)) {
		fprintf(stderr,
			"Inconsistent: detached unreferenced node %d\n",
			i);
	    }
	    if (nodeDetCallbacks) {
		CallbackEvent(E4_CBDETNODE, FindOrCreateNode(i));
	    }
	}
    }
}

void
e4_MetakitStorageImpl::FireEventsForNewlyDetachedVertices()
{
    int i, l, flags;
    bool vertexDetCallbacks;

    vertexDetCallbacks = HasCallbacks(E4_CBDETVERTEX);

    /*
     * We must iterate over all vertices even when there are no
     * callbacks registered, to set the E4_DETACHED flags for
     * newly detached vertices.
     */

    for (i = 0, l = vertices.GetSize(); i < l; i++) {
	flags = (int) pFlags(vertices[i]);
	if ((flags & E4_INUSE) == 0) {
	    continue;
	}

	/*
	 * If this vertex already has been notified of
	 * a detach, don't bother.
	 */

	if ((flags & E4_DETACHNOTIFY) == E4_DETACHNOTIFY) {
	    continue;
	}

	/*
	 * If we found a detached vertex, process it.
	 */

	if ((int) pNodeID(vertices[i]) == E4_NEXTNONE) {
	    flags |= (E4_DETACHED | E4_DETACHNOTIFY);
	    pFlags(vertices[i]) = flags;
	    if (!IsReferencedVertex(i)) {
		fprintf(stderr,
			"Inconsistent: detached unreferenced vertex %d\n",
			i);
	    }
	    if (vertexDetCallbacks) {
		CallbackEvent(E4_CBDETVERTEX, GetVertex(i));
	    }
	}
    }
}

